from __future__ import annotations

import copy
import os
import pathlib
from collections import Counter
from typing import Optional, Union

import crowsetta
import numpy as np
import pandas as pd

from . import constants, files
from .typing import PathLike


def format_from_df(dataset_df: pd.DataFrame) -> str:
    """Get the format of annotations from a dataset,
    given a dataframe representing that dataset.

    Returns string name of annotation format.
    If no annotation format is specified, returns None.
    Raises an error if there are multiple formats.

    Parameters
    ----------
    dataset_df : pandas.DataFrame
        Representing a dataset of vocalizations,
        with column 'annot_format'.

    Returns
    -------
    annot_format : str
        format of annotations for vocalizations.
    """
    annot_format = dataset_df["annot_format"].unique()
    if len(annot_format) == 1:
        annot_format = annot_format.item()
        if (
            annot_format is None
            or annot_format == constants.NO_ANNOTATION_FORMAT
        ):
            return None
    elif len(annot_format) > 1:
        raise ValueError(
            f"unable to load labels for dataset, found multiple annotation formats: {annot_format}"
        )

    return annot_format


def from_df(
    dataset_df: pd.DataFrame, annot_root: str | pathlib.Path | None = None
) -> list[crowsetta.Annotation] | None:
    """Get list of annotations from a dataframe
    representing a dataset.

    If no annotation format is specified for the dataframe
    (in the 'annot_format' column), returns None.

    Parameters
    ----------
    dataset_df : DataFrame
        Dataframe representing a dataset of vocalizations,
        with columns 'annot_format' and 'annot_path'.
    annot_root : str or pathlib.Path, optional
        Path to root of directory where annotation files are located.
        If specified, then paths in the DataFrame from the 'annot_path' column
        are constructed relative to ``annot_root``.
        Default is None, in which case 'annot_paths' are used directly,
        as if they were absolute paths.

    Returns
    -------
    annots : list
        of annotations for each row in the dataframe,
        represented as crowsetta.Annotation instances.

    Notes
    -----
    This function encapsulates logic for handling different types of
    annotations; it determines whether each row has a separate annotation file,
    or if instead there is a single annotation file associated with all rows.
    If the latter, then the function opens that file and makes sure that
    each row from the dataframe can be paired with an annotation
    (using :func:`vak.annotation.map_annotated_to_annot`).
    If instead there is a unique annotation file per row in the dataframe,
    the format of the annotation files is determined with
    :func:`vak.annotation.format_from_df` and then each file is opened
    with :module:`crowsetta` -- in other words, we assume the mapping
    was already done when preparing the dataset, and that each row contains
    an annotation file paired with the file it annotates.
    """
    if annot_root:
        annot_root = pathlib.Path(annot_root)
        if not annot_root.exists() or not annot_root.is_dir():
            raise NotADirectoryError(
                f"`annot_root` not found or not recognized as a directory: {annot_root}"
            )

    annot_format = format_from_df(dataset_df)
    if annot_format is None:
        return None

    scribe = crowsetta.Transcriber(format=annot_format)

    if len(dataset_df["annot_path"].unique()) == 1:
        # --> there is a single annotation file associated with all rows
        # this can be true in two different cases:
        # (1) many rows, all have the same file
        # (2) only one row, so there's only one annotation file (which may contain annotation for multiple source files)
        annot_path = dataset_df["annot_path"].unique().item()
        if annot_root:
            annot_path = annot_root / annot_path
        annots = scribe.from_file(annot_path).to_annot()

        # as long as we have at least as many annotations as there are rows in the dataframe
        if (
            isinstance(annots, list) and len(annots) >= len(dataset_df)
        ) or (  # case 1
            isinstance(annots, crowsetta.Annotation) and len(dataset_df) == 1
        ):  # case 2
            if isinstance(annots, crowsetta.Annotation):
                annots = [
                    annots
                ]  # wrap in list for map_annotated_to_annot to iterate over it
            # then we can try and map those annotations to the rows
            audio_annot_map = map_annotated_to_annot(
                dataset_df["audio_path"].values, annots, annot_format
            )
            # sort by row of dataframe
            annots = [
                audio_annot_map[audio_path]
                for audio_path in dataset_df["audio_path"].values
            ]

        else:
            raise ValueError(
                "unable to load labels from dataframe; found a single annotation file associated with all "
                "rows in dataframe, but loading it did not return a list of annotations for each row.\n"
                f"Single annotation file: {annot_path}\n"
                f"Loading it returned a {type(annots)}."
            )

    elif len(dataset_df["annot_path"].unique()) == len(dataset_df):
        # --> there is a unique annotation file (path) for each row, iterate over them to get labels from each
        annot_paths = dataset_df["annot_path"].values
        if annot_root:
            annot_paths = [
                annot_root / annot_path for annot_path in annot_paths
            ]
        annots = [
            scribe.from_file(annot_path).to_annot()
            for annot_path in annot_paths
        ]

    else:
        raise ValueError(
            "unable to load labels from dataframe; did not find an annotation file for each row or "
            "a single annotation file associated with all rows."
        )

    return annots


def files_from_dir(annot_dir, annot_format):
    """Get all annotation files of a given format
    from a directory or its sub-directories,
    using the file extension associated with that annotation format.
    """
    if annot_format not in constants.VALID_ANNOT_FORMATS:
        raise ValueError(
            f"specified annotation format, {annot_format} not valid.\n"
            f"Valid formats are: {constants.VALID_ANNOT_FORMATS}"
        )

    format_class = crowsetta.formats.by_name(annot_format)
    # handle the case where an annotation format can have more than one valid extension,
    # e.g., simple-seq has ``('.csv', '.txt')`` as extensions
    ext = None
    if isinstance(format_class.ext, str):
        # NOTE that by convention the `ext` attribute
        # of all Crowsetta annotation format classes
        # begins with a period
        ext = format_class.ext
    elif isinstance(format_class.ext, tuple):
        # then we actually have to determine whether there's any files for either format
        for ext_to_test in format_class.ext:
            if (
                len(sorted(pathlib.Path(annot_dir).glob(f"*{ext_to_test}")))
                > 0
            ):
                ext = ext_to_test
    if ext is None:
        raise ValueError(
            f"Unable to determine which extension to use for format: {annot_format}. "
            f"Used extensions from class `{format_class}`, {format_class.ext}, "
            f"but no files were found with that/those extensions in annot_dir:\n{annot_dir}"
        )

    annot_files = files.from_dir(annot_dir, ext)
    return annot_files


class AudioFilenameNotFoundError(Exception):
    """Error raised when a name of an audio filename
    cannot be found within another filename.

    Raised by ``audio_filename_from_path``
    and ``_map_using_audio_stem_from_path``.
    """


def audio_filename_from_path(path: PathLike, audio_ext: str = None) -> str:
    """Find the name of an audio file within a filename
    by removing extensions until finding an audio extension,
    then return the name of that audio file
    without the extension (i.e., the "stem").

    Removes extensions from a filename recursively,
    by calling `os.path.splitext`,
    until the extension is an audio file format handled by vak.
    Then return the stem, that is,
    the part that precedes the extension.
    Used to match audio, spectrogram,
    and annotation files by their stems.

    Stops after finding audio extensions
    so that it does not remove "extensions"
    that are actually other parts of a filename,
    e.g. a time or data separated by periods.

    Examples
    --------
    >>> audio_filename_from_path('gy6or6_baseline_230312_0808.138.cbin.not.mat')
    'gy6or6_baseline_230312_0808.138'
    >>> audio_filename_from_path('Bird0/spectrograms/0.wav.npz')
    '0'
    >>> audio_filename_from_path('Bird0/Wave/0.wav')
    '0'

    Parameters
    ----------
    path : str, Path
        Path to a file that contains an audio filename in its name.
    audio_ext : str
        Extension corresponding to format of audio file.
        Must be one of ``vak.constants.VALID_AUDIO_FORMATS``.
        Default is None, in which case the function looks
        removes extensions until it finds any valid audio
        format extension.

    Returns
    -------
    stem : str
        Part of filename that precedes audio extension.
    """
    if audio_ext:
        if audio_ext.startswith("."):
            audio_ext = audio_ext[1:]
        if audio_ext not in constants.VALID_AUDIO_FORMATS:
            raise ValueError(
                f"Not a valid extension for audio formats: {audio_ext}\n"
                f"Valid formats are: {constants.VALID_AUDIO_FORMATS}"
            )
        extensions_to_look_for = [audio_ext]
    else:
        extensions_to_look_for = constants.VALID_AUDIO_FORMATS

    name = pathlib.Path(path).name
    stem, ext = os.path.splitext(name)
    ext = ext.replace(".", "").lower()
    while ext not in extensions_to_look_for:
        new_stem, ext = os.path.splitext(stem)
        ext = ext.replace(".", "").lower()
        if new_stem == stem:
            raise AudioFilenameNotFoundError(
                f"Unable to find a valid audio filename in path:\n{path}.\n"
                f"Valid audio file extensions are:\n{constants.VALID_AUDIO_FORMATS}"
            )
        else:
            stem = new_stem
    return stem


class MapUsingNotatedPathError(BaseException):
    """Error raised when :func:`vak.annotation._map_using_notated_path`
    cannot map the filename of an annotation file to the name
    of an annotated file"""

    pass


def _map_using_notated_path(
    annotated_files: list[PathLike],
    annot_list: list[crowsetta.Annotation],
    audio_ext: Optional[str] = None,
) -> dict:
    """Map a :class:`list` of annotated files to a :class:`list`
    of  :class:`crowsetta.Annotation` instances,
    using the ``notated_path`` attribute of the
    :class:`~crowsetta.Annotation`.

    This function assumes that the annotation format
    includes the names of the files that it annotates.
    This is necessarily true for any format that puts
    annotations for multiple annotated files into a single
    annotation file.

    One of two helper functions used by
    :func:`~vak.annotation.map_annotated_to_annot`.

    Parameters
    ----------
    annotated_files : list
        List of paths to the annotated files.
    annot_list : list
        List of ``crowsetta.Annotation`` instances.
    audio_ext : str
        Extension corresponding to audio format.
        Valid extension are listed in
        ``vak.constants.VALID_AUDIO_FORMATS``.
        Default is None, in which case the function
        looks for any valid format.

    Returns
    -------
    annotated_annot_map : dict
        Where each key is path to annotated file, and
        its value is the corresponding ``crowsetta.Annotation``.
    """
    # First check that we don't have duplicate keys that would cause this to fail silently
    keys = []
    for annot in annot_list:
        try:
            stem = audio_filename_from_path(annot.notated_path, audio_ext)
        except AudioFilenameNotFoundError as e:
            # Do this as a loop with a super verbose error
            # instead of e.g. a single-line list comprehension
            # so we can help users troubleshoot,
            # see https://github.com/vocalpy/vak/issues/525
            raise MapUsingNotatedPathError(
                "Unable to find an audio filename in the ``notated_path`` attribute of a ``crowsetta.Annotation``."
                f"The ``notated_path`` attribute was:\n{annot.notated_path}\n"
                f"The annotation was loaded from this path:\n{annot.annot_path}\n"
                f"The full annotation is:\n{annot}"
            ) from e
        keys.append(stem)

    keys_set = set(keys)
    if len(keys_set) < len(keys):
        duplicates = [
            item for item, count in Counter(keys).items() if count > 1
        ]
        raise ValueError(
            f"found multiple annotations with the same audio filename(s): {duplicates}"
        )
    del keys, keys_set
    # ----> make a dict with audio filenames as keys,
    #       so we can look up annotations
    #       by getting the same filename from the annotated files themselves,
    #       and using those as keys.
    audio_filename_annot_map = {
        # NOTE HERE WE GET FILENAMES FROM EACH annot.notated_path,
        # BELOW we get filenames from each annotated_file
        audio_filename_from_path(annot.notated_path): annot
        for annot in annot_list
    }

    # Make a copy of ``annotated_files`` from which
    # we remove files after mapping them to annotation,
    # to validate that function worked,
    # by making sure there are no items left in this copy after the loop.
    # If there is 1:1 mapping then there should be no items left.
    annotated_annot_map = {}
    annotated_files_copy = copy.deepcopy(annotated_files)
    for annotated_file in annotated_files:
        # stem annotated file so we can find audio OR spect files
        # that match with stems from each annot.notated_path;
        # e.g. find '~/path/to/llb3/llb3_0003_2018_04_23_14_18_54.wav.mat' that
        # should match with ``Annotation(notated_path='llb3_0003_2018_04_23_14_18_54.wav')``
        audio_filename_from_annotated_file = audio_filename_from_path(
            annotated_file
        )
        try:
            annot = audio_filename_annot_map[
                audio_filename_from_annotated_file
            ]
        except KeyError as e:
            raise MapUsingNotatedPathError(
                "Could not map an annotation to an annotated file path "
                "using `vak.annotation.audio_filename_from_path` to get "
                "an audio filename from the annotated file path."
                f"The annotated file path:\n{annotated_file} "
                "The audio filename found using `vak.annotation.audio_filename_from_path` "
                f"was:\n{audio_filename_from_annotated_file}"
            ) from e
        annotated_annot_map[annotated_file] = annot
        annotated_files_copy.remove(annotated_file)

    if len(annotated_files_copy) > 0:
        raise MapUsingNotatedPathError(
            "Could not map the following source files to annotations: "
            f"{annotated_files_copy}"
        )

    # we return dict[str: annot] since we will always have paths as strings in DataFrame columns
    # and we want to use those strings to index into this dictionary
    return {str(path): annot for path, annot in annotated_annot_map.items()}


class MapUsingExtensionError(BaseException):
    """Error raised when :func:`vak.annotation._map_using_ext`
    cannot map the filename of an annotation file to the name
    of an annotated file"""

    pass


def _map_using_ext(
    annotated_files: list[PathLike],
    annot_list: list[crowsetta.Annotation],
    annot_format: str,
    method: str,
    annotated_ext: str | None = None,
) -> dict:
    """Map a list of annotated files to a :class:`list` of
    :class:`crowsetta.Annotation` instances,
    by either removing the extension of the annotation format,
    or replacing it with the extension of the annotated file format.

    This function assumes a one-to-one mapping between
    annotation files and the files they annotate.

    and that the name of the annotated file is
    the name of the annotation file with its
    format-specific extension removed,
    e.g., a file in a csv-based format named 'bird1.wav.csv'
    annotates a file named `bird1.wav`.

    One of two helper functions used by
    :func:`~.vak.annotation.map_annotated_to_annot`.

    Parameters
    ----------
    annotated_files : list
        List of paths to the annotated files.
    annot_list : list
        List of ``crowsetta.Annotation`` instances.
    annot_format : str
        String name of annotation format
        Valid names are listed in
        ``vak.constants.VALID_ANNOT_FORMATS``.
    method: str
        The "method" used to determine the annotated
        file name from the annotation file name.
        One of {'remove', 'replace'}.
        Corresponds to either removing the extension
        for the annotation file format, or replacing
        its extension with the extension of the annotated
        format.

    Returns
    -------
    annotated_annot_map : dict
        Where each key is path to annotated file, and
        its value is the corresponding ``crowsetta.Annotation``.
    """
    if method not in {"remove", "replace"}:
        raise ValueError(
            f"`method` must be one of: {{'remove', 'replace'}}, but was: '{method}'"
        )

    annotated_files = [
        pathlib.Path(annotated_file) for annotated_file in annotated_files
    ]

    if method == "replace":
        if annotated_ext is None:
            annotated_ext_set = set(
                [annotated_file.suffix for annotated_file in annotated_files]
            )
            if len(annotated_ext_set) > 1:
                raise ValueError(
                    "Found more than one extension in annotated files, "
                    "unclear which extension to use when mapping to annotations "
                    f"with 'replace' method. Extensions found: {annotated_ext_set}"
                )
            annotated_ext = annotated_ext_set.pop()

    annot_class = crowsetta.formats.by_name(annot_format)

    # ---- make the dict that maps name of annotated files to crowsetta.Annotations
    # We do this using names instead of using the full paths so that this function
    # can be directory agnostic, i.e., we ignore the parent path and just use the filename
    # to do the matching. Currently `vak` assumes at a higher level that annotation files
    # and annotated files exist in the same `data_dir` but I am trying to write this
    # function in a slightly more general way. Not obvious to me if there's a way this could backfire.
    # For this function we assume 1:1 mapping between annotated and annotation files,
    # so they probably need to be unique filenames anyway regardless of what dir they are in?
    annotated_filename_annot_map = {}
    for annot in annot_list:
        annotated_name = None
        if isinstance(annot_class.ext, str):
            # NOTE that by convention the `ext` attribute
            # of all Crowsetta annotation format classes
            # begins with a period
            annotated_name = annot.annot_path.name.replace(annot_class.ext, "")
        elif isinstance(annot_class.ext, tuple):
            # handle the case where an annotation format can have multiple extensions,
            # e.g., ``Format.ext == ('.csv', '.txt')``
            for ext in annot_class.ext:
                if annot.annot_path.name.endswith(ext):
                    annotated_name = annot.annot_path.name.replace(ext, "")
                    break

        if annotated_name is None:
            raise MapUsingExtensionError(
                "Could not determine annotated file from annotation path, "
                f"using extension '{annot_class.ext}' from class '{annot_class.__name__}' "
                f"associated with format '{annot_format}'. "
                f"Annotation path was:\n{annot.annot_path}"
            )

        # NOTE we don't have to do anything else for method=='remove'
        # since we just removed the extension
        if method == "replace":
            annotated_name = annotated_name + annotated_ext

        annotated_filename_annot_map[annotated_name] = annot

    annotated_annot_map = {}  # this is what we will return
    # Make a copy of ``annotated_files`` from which
    # we remove files after mapping them to annotation,
    # to validate that function worked,
    # by making sure there are no items left in this copy after the loop.
    # If there is 1:1 mapping then there should be no items left.
    annotated_files_copy = copy.deepcopy(annotated_files)
    for annotated_file in annotated_files:
        try:
            annot = annotated_filename_annot_map[annotated_file.name]
            annotated_files_copy.remove(annotated_file)
        except KeyError as e:
            raise MapUsingExtensionError(
                f"Did not find an annotation that produced annotated file: {annotated_file}"
            ) from e
        annotated_annot_map[annotated_file] = annot

    if len(annotated_files_copy) > 0:
        raise MapUsingExtensionError(
            "Could not map the following source files to annotations: "
            f"{annotated_files_copy}"
        )
    # we return dict[str: annot] since we will always have paths as strings in DataFrame columns
    # and we want to use those strings to index into this dictionary
    return {str(path): annot for path, annot in annotated_annot_map.items()}


def map_annotated_to_annot(
    annotated_files: Union[list, np.array],
    annot_list: list[crowsetta.Annotation],
    annot_format: str,
    annotated_ext: str | None = None,
) -> dict:
    """Map annotated files,
    i.e. audio or spectrogram files,
    to their corresponding annotations.

    This function implements the three different ways that
    vak can map annotated files to their annotations.
    The first is when a single annotation file contains
    multiple annotations, and so the format by necessity
    must include the file annotated by each annotation.
    The second assumes that the annotated file can be determined
    programmatically by removing the extension from the annotation file,
    e.g. 'bird1.wav.csv' -> 'bird1.wav'.
    The third assumes that the annotated file can be determined
    by replacing the extension of the annotation file
    with the extension of the annotated file,
    e.g. 'bird1.csv' -> 'bird1.wav'.

    Returns a :class:`dict` where each key
    is a path to an annotated file,
    and the value for each key
    is a :class:`crowsetta.Annotation`.

    Mapping is done with two helper functions,
    :func:`~vak.annotation._map_using_notated_path` and
    :func:`~vak.annotation._map_using_ext`.

    The function :func:`~vak.annotation._map_using_notated_path`
    is used for annotation formats that include
    the name of the annotated file.
    The names of these formats (in :module:`crowsetta`) are:
    {'birdsong-recognition-dataset', 'generic-seq', 'yarden'}.

    The other function is are used for all other formats,
    and it assumes a one-to-one mapping from annotation file
    to annotated file.
    It assumes that the name of the annotated file
    can be found by removing the extension of the annotation
    format, e.g., 'bird1.wav.csv` -> 'bird1.wav'.
    The second, that is used if the first fails,
    assumes the name of the annotated file
    can be found by replacing the extension of the annotation
    format with the extension of the annotated files.

    Parameters
    ----------
    annotated_files : list
        Of paths to audio or spectrogram files.
    annot_list : list
        Of Annotations corresponding to files in annotated_files
    annotated_ext : str
        Extension of annotated files.
        Default is None, in which case this function will
        look for extensions of any valid audio format
        (listed as ``vak.constants.VALID_AUDIO_FORMAT``).
        Specifying the format provides a slight speed up.

    Notes
    -----
    For more detail, please see
    the page on file naming conventions in the
    reference section of the documentation:
    https://vak.readthedocs.io/en/latest/reference/filenames.html
    """
    if isinstance(
        annotated_files, np.ndarray
    ):  # e.g., vak DataFrame['spect_path'].values
        annotated_files = annotated_files.tolist()

    if annot_format in (
        "birdsong-recognition-dataset",
        "yarden",
        "generic-seq",
    ):
        annotated_annot_map = _map_using_notated_path(
            annotated_files, annot_list
        )
    else:
        try:
            annotated_annot_map = _map_using_ext(
                annotated_files, annot_list, annot_format, method="remove"
            )
        except MapUsingExtensionError:
            try:
                annotated_annot_map = _map_using_ext(
                    annotated_files,
                    annot_list,
                    annot_format,
                    method="replace",
                    annotated_ext=annotated_ext,
                )
            except MapUsingExtensionError as e:
                raise ValueError(
                    "Could not map annotated files to annotations.\n"
                    "Please see this section in the `vak` documentation:\n"
                    "https://vak.readthedocs.io/en/latest/howto/howto_prep_annotate.html"
                    "#how-does-vak-know-which-annotations-go-with-which-annotated-files"
                ) from e

    return annotated_annot_map


def has_unlabeled(annot: crowsetta.Annotation, duration: float) -> bool:
    """Returns ``True`` if an annotated sequence has unlabeled segments.

    Tests whether an instance of ``crowsetta.Annotation.seq`` has
    intervals between the annotated segments with a non-zero duration,
    or any unannotated periods before or after the annotated segments.

    Parameters
    ----------
    annot : crowsetta.Annotation
        A :class:`crowsetta.Annotation` with a ``seq`` attribute
        (that is a :class:`crowsetta.Sequence`).
    duration : float
        Total duration of the vocalization
        that is annotated by ``annot``.
        Needed to determine whether the duration
        is greater than the time
        of the last offset in the annotated segments.

    Returns
    -------
    has_unlabeled : bool
        If True, there are unlabeled periods
        in the vocalization annotated by ``annot``.
    """
    if duration <= 0:
        raise ValueError(
            f"Duration less than or equal to zero passed to ``has_unlabeled``.\n"
            f"Value for ``duration``: {duration}.\nValue for ``annot``: {annot}"
        )
    if duration > 0 and len(annot.seq.segments) < 1:
        # Handle edge case where there are no annotated segments in annotation file
        # See https://github.com/vocalpy/vak/issues/378
        return True
    has_unlabeled_intervals = np.any(
        (annot.seq.onsets_s[1:] - annot.seq.offsets_s[:-1]) > 0.0
    )
    has_unlabeled_before_first_onset = annot.seq.onsets_s[0] > 0.0
    has_unlabeled_after_last_offset = duration - annot.seq.offsets_s[-1] > 0.0
    return (
        has_unlabeled_intervals
        or has_unlabeled_before_first_onset
        or has_unlabeled_after_last_offset
    )



"""Helper functions for frame classification dataset prep."""

from __future__ import annotations

import collections
import copy
import logging
import pathlib
import shutil

import attrs
import crowsetta
import dask.bag as db
import numpy as np
import pandas as pd
from dask.diagnostics import ProgressBar

from ... import common, datapipes, transforms
from .. import constants as prep_constants

logger = logging.getLogger(__name__)


def argsort_by_label_freq(annots: list[crowsetta.Annotation]) -> list[int]:
    """Returns indices to sort a list of annotations
     in order of more frequently appearing labels,
     i.e., the first annotation will have the label
     that appears least frequently and the last annotation
     will have the label that appears most frequently.

    Used to sort a dataframe representing a dataset of annotated audio
    or spectrograms before cropping that dataset to a specified duration,
    so that it's less likely that cropping will remove all occurrences
    of any label class from the total dataset.

     Parameters
     ----------
     annots: list
         List of :class:`crowsetta.Annotation` instances.

     Returns
     -------
     sort_inds: list
         Integer values to sort ``annots``.
    """
    all_labels = [lbl for annot in annots for lbl in annot.seq.labels]
    label_counts = collections.Counter(all_labels)

    sort_inds = []
    # make indices ahead of time so they stay constant as we remove things from the list
    ind_annot_tuples = list(enumerate(copy.deepcopy(annots)))
    for label, _ in reversed(label_counts.most_common()):
        # next line, [:] to make a temporary copy to avoid remove bug
        for ind_annot_tuple in ind_annot_tuples[:]:
            ind, annot = ind_annot_tuple
            if label in annot.seq.labels.tolist():
                sort_inds.append(ind)
                ind_annot_tuples.remove(ind_annot_tuple)

    # make sure we got all source_paths + annots
    if len(ind_annot_tuples) > 0:
        for ind_annot_tuple in ind_annot_tuples:
            ind, annot = ind_annot_tuple
            sort_inds.append(ind)
            ind_annot_tuples.remove(ind_annot_tuple)

    if len(ind_annot_tuples) > 0:
        raise ValueError(
            "Not all ``annots`` were used in sorting."
            f"Left over (with indices from list): {ind_annot_tuples}"
        )

    if not (sorted(sort_inds) == list(range(len(annots)))):
        raise ValueError(
            "sorted(sort_inds) does not equal range(len(annots)):"
            f"sort_inds: {sort_inds}\nrange(len(annots)): {list(range(len(annots)))}"
        )

    return sort_inds


@attrs.define(frozen=True)
class Sample:
    """Dataclass representing one sample
    in a frame classification dataset.

    Used to add paths for arrays from the sample
    to a ``dataset_df``, and to build
    the ``sample_ids`` vector and ``inds_in_sample`` vector
    for the entire dataset.

    Attributes
    ----------
    source_id : int
        Integer ID number used for sorting.
    frames_path : str
        The path to the input to the model
        :math:`x` after it has been moved,
        copied, or created from a ``source_path``.
        Path will be written relative to ``dataset_path``.
        We preserve the original paths as metadata,
        and consider the files in the split to contain
        frames, regardless of the source domain
        of the data.
    frame_labels_npy_path : str
        Path to frame labels,
        relative to ``dataset_path``.
    sample_id_vec : numpy.ndarray
        Sample ID vector for this sample.
    inds_in_sample_vec : numpy.ndarray
        Indices within sample.
    """

    source_id: int = attrs.field()
    source_path: str
    frame_labels_npy_path: str
    sample_id_vec: np.ndarray
    inds_in_sample_vec: np.ndarray


def make_splits(
    dataset_df: pd.DataFrame,
    dataset_path: str | pathlib.Path,
    input_type: str,
    purpose: str,
    labelmap: dict,
    audio_format: str | None = None,
    spect_key: str = "s",
    timebins_key: str = "t",
    freqbins_key: str = "f",
    background_label: str = common.constants.DEFAULT_BACKGROUND_LABEL,
) -> pd.DataFrame:
    r"""Make each split of a frame classification dataset.

    This function takes a :class:`pandas.Dataframe` returned by
    :func:`vak.prep.spectrogram_dataset.prep_spectrogram_dataset`
    or :func:`vak.prep.audio_dataset.prep_audio_dataset`,
    after it has been assigned a `'split'` column,
    and then copies, moves, or generates the required files
    as appropriate for each split.

    For each unique `'split'` in the :class:`pandas.Dataframe`,
    a directory is made inside ``dataset_path``.
    At a high level, all files needed for working with that split
    will be in that directory
    E.g., the ``train`` directory inside ``dataset_path``
    would have all the files for every row in ``dataset_df``
    for which ``dataset_df['split'] == 'train'``.

    The inputs to the neural network model
    are moved or copied into the split directory,
    or generated if necessary.
    If the ``input_type`` is `'audio'`,
    then the audio files are copied from their original directory.
    If the ``input_type`` is `'spect'`,
    and the spectrogram files are already
    in ``dataset_path``, they are moved into the split directory
    (under the assumption they were generated
    by ``vak.prep.spectrogram_dataset.audio_helper``).
    If they are npz files, but they are not in ``dataset_path``,
    then they are validated to make sure they have the appropriate keys,
    and then copied into the split directory.
    This could be the case if the files were generated
    by another program.
    If they are mat files, they will be converted to npz
    with the default keys for arrays,
    and then saved in a new npz file in the split directory.
    This step is required so that all dataset
    prepared by :mod:`vak` are in a "normalized" or
    "canonicalized" format.

    In addition to copying or moving the audio or spectrogram
    files that are inputs to the neural network model,
    other npy files are made for each split
    and saved in the corresponding directory.
    This function creates one npy file for each row in ``dataset_df``.
    It has the extension '.frame_labels.npy', and contains a vector
    where each element is the target label that
    the network should predict for the corresponding frame.
    Taken together, the audio or spectrogram file in each row
    along with its corresponding frame labels are the data
    for each sample :math:`(x, y)` in the dataset,
    where :math:`x_t` supplies the "frames", and :math:`y_t` is the frame labels.

    This function also creates two additional npy files for each split.
    These npy files are "indexing" vectors that
    are used by :class:`vak.datasets.frame_classification.TrainDatapipe`
    and :class:`vak.datasets.frame_classification.InferDatapipe`.
    These vectors make it possible to work with files,
    to avoid loading the entire dataset into memory,
    and to avoid working with memory-mapped arrays.
    The first is the ``sample_ids`` vector,
    that represents the "ID" of any sample :math:`(x, y)` in the split.
    We use these IDs to load the array files corresponding to the samples.
    For a split with :math:`m` samples, this will be an array of length :math:`T`,
    the total number of frames across all samples,
    with elements :math:`i \in (0, 1, ..., m - 1)`
    indicating which frames correspond to which sample :math:`m_i`:
    :math:`(0, 0, 0, ..., 1, 1, ..., m - 1, m -1)`.
    The second vector is the ``inds_in_sample`` vector.
    This vector is the same length as ``sample_ids``, but its values represent
    the indices of frames within each sample :math:`x_t`.
    For a data set with :math:`T` total frames across all samples,
    where :math:`t_i` indicates the number of frames in each :math:`x_i`,
    this vector will look like :math:`(0, 1, ..., t_0, 0, 1, ..., t_1, ... t_m)`.

    Parameters
    ----------
    dataset_df : pandas.DataFrame
        A ``pandas.DataFrame`` returned by :func:`vak.io.dataframe.from_files`
        with a ``'split'`` column added, as a result of calling
        :func:`vak.io.dataframe.from_files` or because it was added "manually"
        by calling :func:`vak.core.prep.prep_helper.add_split_col` (as is done
        for 'predict' when the entire ``DataFrame`` belongs to this
        "split").
    dataset_path : pathlib.Path
        Path to directory that represents dataset.
    input_type : str
        The type of input to the neural network model.
        One of {'audio', 'spect'}.
    purpose: str
        A string indicating what the dataset will be used for.
        One of {'train', 'eval', 'predict', 'learncurve'}.
        Determined by :func:`vak.core.prep.prep`
        using the TOML configuration file.
    labelmap : dict
        A :class:`dict` that maps a set of human-readable
        string labels to the integer classes predicted by a neural
        network model. As returned by :func:`vak.labels.to_map`.
    audio_format : str
        A :class:`string` representing the format of audio files.
        One of :const:`vak.common.constants.VALID_AUDIO_FORMATS`.
    spect_key : str
        Key for accessing spectrogram in files. Default is 's'.
    timebins_key : str
        Key for accessing vector of time bins in files. Default is 't'.
    freqbins_key : str
        key for accessing vector of frequency bins in files. Default is 'f'.
    background_label: str, optional
        The string label applied to segments belonging to the
        background class.
        Default is
        :const:`vak.common.constants.DEFAULT_BACKGROUND_LABEL`.

    Returns
    -------
    dataset_df_out : pandas.DataFrame
        The ``dataset_df`` with splits sorted by increasing frequency
        of labels (see :func:`~vak.prep.frame_classification.dataset_arrays`),
        and with columns added containing the npy files for each row.
    """
    if input_type not in prep_constants.INPUT_TYPES:
        raise ValueError(
            f"``input_type`` must be one of: {prep_constants.INPUT_TYPES}\n"
            f"Value for ``input_type`` was: {input_type}"
        )

    if input_type == "audio" and audio_format is None:
        raise ValueError(
            "Value for `input_type` was 'audio' but `audio_format` is None. "
            "Please specify the audio format."
        )

    dataset_df_out = []
    splits = [
        split
        for split in sorted(dataset_df.split.dropna().unique())
        if split != "None"
    ]
    for split in splits:
        logger.info(f"Making split for dataset: {split}")
        split_subdir = dataset_path / split
        split_subdir.mkdir()

        split_df = dataset_df[dataset_df.split == split].copy()

        if purpose != "predict":
            annots = common.annotation.from_df(split_df)
        else:
            annots = None

        if annots:
            sort_inds = argsort_by_label_freq(annots)
            split_df["sort_inds"] = sort_inds
            split_df = split_df.sort_values(by="sort_inds").drop(
                columns="sort_inds"
            )

        if input_type == "audio":
            source_paths = split_df["audio_path"].values
        elif input_type == "spect":
            source_paths = split_df["spect_path"].values
        else:
            raise ValueError(f"Invalid ``input_type``: {input_type}")
        source_paths = [
            pathlib.Path(source_path) for source_path in source_paths
        ]

        # we get annots again, *after* sorting the dataframe
        if purpose != "predict":
            annots = common.annotation.from_df(split_df)
        else:
            annots = None

        def _save_dataset_arrays_and_return_index_arrays(
            source_id_path_annot_tup,
        ):
            """Function we use with dask to parallelize

            Defined in-line so variables are in scope
            """
            source_id, source_path, annot = source_id_path_annot_tup

            if input_type == "audio":
                # we always copy audio to the split directory, to avoid damaging source data
                frames_path = shutil.copy(source_path, split_subdir)
                # after copying, we load frames to compute frame labels
                frames, samplefreq = common.constants.AUDIO_FORMAT_FUNC_MAP[
                    audio_format
                ](source_path)
                if (
                    audio_format == "cbin"
                ):  # convert to ~wav, from int16 to float64damage
                    frames = frames.astype(np.float64) / 32768.0
                if annot:
                    frame_times = np.arange(frames.shape[-1]) / samplefreq
            elif input_type == "spect":
                if source_path.suffix.endswith("mat"):
                    spect_dict = common.files.spect.load(source_path, "mat")
                    # convert to .npz and save in spect_output_dir
                    spect_dict_npz = {
                        "s": spect_dict[spect_key],
                        "t": spect_dict[timebins_key],
                        "f": spect_dict[freqbins_key],
                    }
                    frames_path = split_subdir / (source_path.stem + ".npz")
                    np.savez(frames_path, **spect_dict_npz)
                elif source_path.suffix.endswith("npz"):
                    spect_dict = common.files.spect.load(source_path, "npz")
                    if source_path.is_relative_to(dataset_path):
                        # it's already in dataset_path, we just move it into the split
                        frames_path = shutil.move(source_path, split_subdir)
                    else:
                        # it's somewhere else we copy it to be safe
                        if not all(
                            [key in spect_dict for key in ("s", "t", "f")]
                        ):
                            raise ValueError(
                                f"The following spectrogram file did not have valid keys: {source_path}\n."
                                f"All npz files should have keys 's', 't', 'f' corresponding to the spectrogram,"
                                f"the frequencies vector, and the time vector."
                            )
                        frames_path = shutil.copy(source_path, split_subdir)
                frames = spect_dict[spect_key]
                if annot:
                    frame_times = spect_dict[timebins_key]

            n_frames = frames.shape[-1]
            sample_id_vec = np.ones((n_frames,)).astype(np.int32) * source_id
            inds_in_sample_vec = np.arange(n_frames)

            # add to frame labels
            if annot:
                lbls_int = [labelmap[lbl] for lbl in annot.seq.labels]
                frame_labels = transforms.frame_labels.from_segments(
                    lbls_int,
                    annot.seq.onsets_s,
                    annot.seq.offsets_s,
                    frame_times,
                    background_label=labelmap[background_label],
                )
                frame_labels_npy_path = split_subdir / (
                    source_path.stem
                    + datapipes.frame_classification.constants.MULTI_FRAME_LABELS_EXT
                )
                np.save(frame_labels_npy_path, frame_labels)
                frame_labels_npy_path = str(
                    # make sure we save path in csv as relative to dataset root
                    frame_labels_npy_path.relative_to(dataset_path)
                )
            else:
                frame_labels_npy_path = None

            # Rewrite ``frames_path`` as relative to root
            # because all functions and classes downstream expect this
            frames_path = pathlib.Path(frames_path).relative_to(dataset_path)

            return Sample(
                source_id,
                frames_path,
                frame_labels_npy_path,
                sample_id_vec,
                inds_in_sample_vec,
            )

        # ---- make npy files for this split, parallelized with dask
        # using nested function just defined
        if annots:
            source_path_annot_tups = [
                (source_id, source_path, annot)
                for source_id, (source_path, annot) in enumerate(
                    zip(source_paths, annots)
                )
            ]
        else:
            source_path_annot_tups = [
                (source_id, source_path, None)
                for source_id, source_path in enumerate(source_paths)
            ]

        source_path_annot_bag = db.from_sequence(source_path_annot_tups)
        with ProgressBar():
            samples = list(
                source_path_annot_bag.map(
                    _save_dataset_arrays_and_return_index_arrays
                )
            )
        samples = sorted(samples, key=lambda sample: sample.source_id)

        # ---- save indexing vectors in split directory
        sample_id_vec = np.concatenate(
            list(sample.sample_id_vec for sample in samples)
        )
        np.save(
            split_subdir
            / datapipes.frame_classification.constants.SAMPLE_IDS_ARRAY_FILENAME,
            sample_id_vec,
        )
        inds_in_sample_vec = np.concatenate(
            list(sample.inds_in_sample_vec for sample in samples)
        )
        np.save(
            split_subdir
            / datapipes.frame_classification.constants.INDS_IN_SAMPLE_ARRAY_FILENAME,
            inds_in_sample_vec,
        )

        # We convert `frames_paths` back to string
        # (just in case they are pathlib.Paths) before adding back to dataframe.
        # Note that these are all in split dirs, written relative to ``dataset_path``.
        frames_paths = [str(sample.source_path) for sample in samples]
        split_df[
            datapipes.frame_classification.constants.FRAMES_PATH_COL_NAME
        ] = frames_paths

        frame_labels_npy_paths = [
            (
                sample.frame_labels_npy_path
                if isinstance(sample.frame_labels_npy_path, str)
                else None
            )
            for sample in samples
        ]
        split_df[
            datapipes.frame_classification.constants.MULTI_FRAME_LABELS_PATH_COL_NAME
        ] = frame_labels_npy_paths
        dataset_df_out.append(split_df)

    # ---- clean up
    # Remove any spect npz files that were *not* added to a split
    spect_npz_files_not_in_split = sorted(
        dataset_path.glob(f"*{common.constants.SPECT_NPZ_EXTENSION}")
    )
    if len(spect_npz_files_not_in_split) > 0:
        for spect_npz_file in spect_npz_files_not_in_split:
            spect_npz_file.unlink()

    # we reset the entire index across all splits, instead of repeating indices,
    # and we set drop=False because we don't want to add a new column 'index' or 'level_0'
    dataset_df_out = pd.concat(dataset_df_out).reset_index(drop=True)
    return dataset_df_out



"""Class representing BioSoundSegBench dataset."""

from __future__ import annotations

import json
import pathlib
from typing import TYPE_CHECKING, Callable, Literal

import numpy as np
import pandas as pd
import torch
import torchvision.transforms
from attrs import define

from .. import common, datapipes, transforms

if TYPE_CHECKING:
    from ..transforms import FramesStandardizer


VALID_TARGET_TYPES = (
    "boundary_frame_labels",
    "multi_frame_labels",
    "binary_frame_labels",
    ("boundary_frame_labels", "multi_frame_labels"),
    "None",
)


FRAMES_PATH_COL_NAME = "frames_path"
MULTI_FRAME_LABELS_PATH_COL_NAME = "multi_frame_labels_path"
BINARY_FRAME_LABELS_PATH_COL_NAME = "binary_frame_labels_path"
BOUNDARY_FRAME_LABELS_PATH_COL_NAME = "boundary_frame_labels_path"


@define
class SampleIDVectorPaths:
    train: pathlib.Path
    val: pathlib.Path
    test: pathlib.Path


@define
class IndsInSampleVectorPaths:
    train: pathlib.Path
    val: pathlib.Path
    test: pathlib.Path


@define
class SplitsMetadata:
    """Class that represents metadata about dataset splits
    in the BioSoundSegBench dataset, loaded from a json file"""

    splits_csv_path: pathlib.Path
    sample_id_vector_paths: SampleIDVectorPaths
    inds_in_sample_vector_paths: IndsInSampleVectorPaths

    @classmethod
    def from_paths(cls, json_path, dataset_path):
        json_path = pathlib.Path(json_path)
        with json_path.open("r") as fp:
            splits_json = json.load(fp)

        dataset_path = pathlib.Path(dataset_path)
        if not dataset_path.exists() or not dataset_path.is_dir():
            raise NotADirectoryError(
                f"`dataset_path` not found or not a directory: {dataset_path}"
            )

        splits_csv_path = pathlib.Path(
            dataset_path / splits_json["splits_csv_path"]
        )
        if not splits_csv_path.exists():
            raise FileNotFoundError(
                f"`splits_csv_path` not found: {splits_csv_path}"
            )

        sample_id_vector_paths = {
            split: dataset_path / path
            for split, path in splits_json["sample_id_vec_path"].items()
        }
        for split, vec_path in sample_id_vector_paths.items():
            if not vec_path.exists():
                raise FileNotFoundError(
                    f"`sample_id_vector_path` for split '{split}' not found: {vec_path}"
                )
        sample_id_vector_paths = SampleIDVectorPaths(**sample_id_vector_paths)

        inds_in_sample_vector_paths = {
            split: dataset_path / path
            for split, path in splits_json["inds_in_sample_vec_path"].items()
        }
        for split, vec_path in inds_in_sample_vector_paths.items():
            if not vec_path.exists():
                raise FileNotFoundError(
                    f"`inds_in_sample_vec_path` for split '{split}' not found: {vec_path}"
                )
        inds_in_sample_vector_paths = IndsInSampleVectorPaths(
            **inds_in_sample_vector_paths
        )

        return cls(
            splits_csv_path,
            sample_id_vector_paths,
            inds_in_sample_vector_paths,
        )


@define
class TrainingReplicateMetadata:
    """Class representing metadata for a
    pre-defined training replicate
    in the BioSoundSegBench dataset.
    """

    biosound_group: str
    id: str | None
    frame_dur: float
    unit: str
    data_source: str | None
    train_dur: float
    replicate_num: int


def metadata_from_splits_json_path(
    splits_json_path: pathlib.Path, datset_path: pathlib.Path
) -> TrainingReplicateMetadata:
    name = splits_json_path.name
    try:
        (
            biosound_group,
            unit,
            id_,
            frame_dur_1st_half,
            frame_dur_2nd_half,
            data_source,
            train_dur_1st_half,
            train_dur_2nd_half,
            replicate_num,
            _,
            _,
        ) = name.split(".")
    # Human-Speech doesn't have ID or data source in filename
    # so it will raise a ValueError
    except ValueError:
        name = splits_json_path.name
        (
            biosound_group,
            unit,
            frame_dur_1st_half,
            frame_dur_2nd_half,
            train_dur_1st_half,
            train_dur_2nd_half,
            replicate_num,
            _,
            _,
        ) = name.split(".")
        id_ = None
        data_source = None
    if id_ is not None:
        id_ = id_.split("-")[-1]
    frame_dur = float(
        frame_dur_1st_half.split("-")[-1]
        + "."
        + frame_dur_2nd_half.split("-")[0]
    )
    train_dur = float(
        train_dur_1st_half.split("-")[-1]
        + "."
        + train_dur_2nd_half.split("-")[0]
    )
    replicate_num = int(replicate_num.split("-")[-1])
    return TrainingReplicateMetadata(
        biosound_group,
        id_,
        frame_dur,
        unit,
        data_source,
        train_dur,
        replicate_num,
    )


class TrainItemTransform:
    """Default transform used when training frame classification models
    with :class:`BioSoundSegBench` dataset."""

    def __init__(
        self,
        frames_standardizer: FramesStandardizer | None = None,
    ):
        from ..transforms import FramesStandardizer  # avoid circular import

        if frames_standardizer is not None:
            if isinstance(frames_standardizer, FramesStandardizer):
                frames_transform = [frames_standardizer]
            else:
                raise TypeError(
                    f"invalid type for frames_standardizer: {type(frames_standardizer)}. "
                    "Should be an instance of vak.transforms.StandardizeSpect"
                )
        else:
            frames_transform = []
        # add as an attribute on self so that high-level functions can save this class as needed
        self.frames_standardizer = frames_standardizer

        frames_transform.extend(
            [
                transforms.ToFloatTensor(),
                transforms.AddChannel(),
            ]
        )
        self.frames_transform = torchvision.transforms.Compose(
            frames_transform
        )
        self.frame_labels_transform = transforms.ToLongTensor()

    def __call__(
        self,
        frames: torch.Tensor,
        multi_frame_labels: torch.Tensor | None = None,
        binary_frame_labels: torch.Tensor | None = None,
        boundary_frame_labels: torch.Tensor | None = None,
    ) -> dict:
        frames = self.frames_transform(frames)
        item = {
            "frames": frames,
        }
        if multi_frame_labels is not None:
            item["multi_frame_labels"] = self.frame_labels_transform(
                multi_frame_labels
            )

        if binary_frame_labels is not None:
            item["binary_frame_labels"] = self.frame_labels_transform(
                binary_frame_labels
            )

        if boundary_frame_labels is not None:
            item["boundary_frame_labels"] = self.frame_labels_transform(
                boundary_frame_labels
            )

        return item


class InferItemTransform:
    """Default transform used when running inference on classification models
    with :class:`BioSoundSegBench` dataset, for evaluation or to generate new predictions.

    Returned item includes frames reshaped into a stack of windows,
    with padded added to make reshaping possible.
    Any `frame_labels` are not padded and reshaped,
    but are converted to :class:`torch.LongTensor`.
    If return_padding_mask is True, item includes 'padding_mask' that
    can be used to crop off any predictions made on the padding.

    Attributes
    ----------
    frames_standardizer : vak.transforms.FramesStandardizer
        instance that has already been fit to dataset, using fit_df method.
        Default is None, in which case no standardization transform is applied.
    window_size : int
        width of window in number of elements. Argument to PadToWindow transform.
    frames_padval : float
        Value to pad frames with. Added to end of array, the "right side".
        Argument to PadToWindow transform. Default is 0.0.
    frame_labels_padval : int
        Value to pad frame labels vector with. Added to the end of the array.
        Argument to PadToWindow transform. Default is -1.
        Used with ``ignore_index`` argument of :mod:`torch.nn.CrossEntropyLoss`.
    return_padding_mask : bool
        if True, the dictionary returned by ItemTransform classes will include
        a boolean vector to use for cropping back down to size before padding.
        padding_mask has size equal to width of padded array, i.e. original size
        plus padding at the end, and has values of 1 where
        columns in padded are from the original array,
        and values of 0 where columns were added for padding.
    """

    def __init__(
        self,
        window_size,
        frames_standardizer=None,
        frames_padval=0.0,
        frame_labels_padval=-1,
        return_padding_mask=True,
        channel_dim=1,
    ):
        from ..transforms import FramesStandardizer  # avoid circular import

        self.window_size = window_size
        self.frames_padval = frames_padval
        self.frame_labels_padval = frame_labels_padval
        self.return_padding_mask = return_padding_mask
        self.channel_dim = channel_dim

        if frames_standardizer is not None:
            if not isinstance(frames_standardizer, FramesStandardizer):
                raise TypeError(
                    f"Invalid type for frames_standardizer: {type(frames_standardizer)}. "
                    "Should be an instance of vak.transforms.FramesStandardizer"
                )
        # add as an attribute on self to use inside __call__
        # *and* so that high-level functions can save this class as needed
        self.frames_standardizer = frames_standardizer

        self.pad_to_window = transforms.PadToWindow(
            window_size, frames_padval, return_padding_mask=return_padding_mask
        )

        self.frames_transform_after_pad = torchvision.transforms.Compose(
            [
                transforms.ViewAsWindowBatch(window_size),
                transforms.ToFloatTensor(),
                # below, add channel at first dimension because windows become batch
                transforms.AddChannel(channel_dim=channel_dim),
            ]
        )

        self.frame_labels_padval = frame_labels_padval
        self.frame_labels_transform = transforms.ToLongTensor()

    def __call__(
        self,
        frames: torch.Tensor,
        multi_frame_labels: torch.Tensor | None = None,
        binary_frame_labels: torch.Tensor | None = None,
        boundary_frame_labels: torch.Tensor | None = None,
        frames_path=None,
    ) -> dict:
        if self.frames_standardizer:
            frames = self.frames_standardizer(frames)

        if self.pad_to_window.return_padding_mask:
            frames, padding_mask = self.pad_to_window(frames)
        else:
            frames = self.pad_to_window(frames)
            padding_mask = None
        frames = self.frames_transform_after_pad(frames)

        item = {
            "frames": frames,
        }

        if multi_frame_labels is not None:
            item["multi_frame_labels"] = self.frame_labels_transform(
                multi_frame_labels
            )

        if binary_frame_labels is not None:
            item["binary_frame_labels"] = self.frame_labels_transform(
                binary_frame_labels
            )

        if boundary_frame_labels is not None:
            item["boundary_frame_labels"] = self.frame_labels_transform(
                boundary_frame_labels
            )

        if padding_mask is not None:
            item["padding_mask"] = padding_mask

        if frames_path is not None:
            # make sure frames_path is a str, not a pathlib.Path
            item["frames_path"] = str(frames_path)

        return item


class BioSoundSegBench:
    """Class representing BioSoundSegBench dataset."""

    def __init__(
        self,
        dataset_path: str | pathlib.Path,
        splits_path: str | pathlib.Path,
        split: Literal["train", "val", "test"],
        window_size: int,
        target_type: str | list[str] | tuple[str] | None = None,
        stride: int = 1,
        standardize_frames: bool = False,
        frames_standardizer: transforms.FramesStandardizer | None = None,
        frames_padval: float = 0.0,
        frame_labels_padval: int = -1,
        return_padding_mask: bool = False,
        return_frames_path: bool = False,
        item_transform: Callable | None = None,
    ):
        """BioSoundSegBench dataset."""
        # ---- validate args, roughly in order
        dataset_path = pathlib.Path(dataset_path)
        if not dataset_path.exists() or not dataset_path.is_dir():
            raise NotADirectoryError(
                f"`dataset_path` for dataset not found, or not a directory: {dataset_path}"
            )
        self.dataset_path = dataset_path
        if split not in common.constants.VALID_SPLITS:
            raise ValueError(
                f"Invalid split name: {split}\n"
                f"Valid splits are: {common.constants.VALID_SPLITS}"
            )

        splits_path = pathlib.Path(splits_path)
        if not splits_path.exists():
            tmp_splits_path = (
                dataset_path / "splits" / "splits-jsons" / splits_path
            )
            if not tmp_splits_path.exists():
                raise FileNotFoundError(
                    f"Did not find `splits_path` using either absolute path ({splits_path})"
                    f"or relative to `dataset_path` ({tmp_splits_path})"
                )
            # if tmp_splits_path *does* exist, replace splits_path with it
            splits_path = tmp_splits_path
        self.splits_path = splits_path
        self.splits_metadata = SplitsMetadata.from_paths(
            json_path=splits_path, dataset_path=dataset_path
        )

        if target_type is None and split != "predict":
            raise ValueError(
                f"Must specify `target_type` if split is '{split}', but "
                "`target_type` is None. `target_type` can only be None if split is 'predict'."
            )
        if target_type is None:
            target_type = "None"
        if not isinstance(target_type, (str, list, tuple)):
            raise TypeError(
                f"`target_type` must be string or sequence of strings but type was: {type(target_type)}\n"
                f"Valid `target_type` arguments are: {VALID_TARGET_TYPES}"
            )
        if isinstance(target_type, (list, tuple)):
            if not all(
                [isinstance(target_type_, str) for target_type_ in target_type]
            ):
                types_in_target_types = set(
                    [type(target_type_) for target_type_ in target_type]
                )
                raise TypeError(
                    "A list or tuple of `target_type` must be all strings, "
                    f"but found the following types: {types_in_target_types}\n"
                    f"`target_type` was: {target_type}\n"
                    f"Valid `target_type` arguments are: {VALID_TARGET_TYPES}"
                )
            # alphabetically sort list or tuple, and make sure it's a tuple
            target_type = tuple(sorted(target_type))
        if target_type not in VALID_TARGET_TYPES:
            raise ValueError(
                f"Invalid `target_type`: {target_type}. "
                f"Valid target types are: {VALID_TARGET_TYPES}"
            )
        if isinstance(target_type, str):
            # make single str a tuple so we can do ``if 'some target' in self.target_type``
            target_type = (target_type,)
        self.target_type = target_type

        # this is a bit convoluted: we are setting metadata, to set frame dur,
        # to be able to compute duration in property below
        self.training_replicate_metadata = metadata_from_splits_json_path(
            self.splits_path, self.dataset_path
        )
        self.frame_dur = (
            self.training_replicate_metadata.frame_dur * 1e-3
        )  # convert from ms to s!

        if "multi_frame_labels" in target_type:
            labelmaps_json_path = self.dataset_path / "labelmaps.json"
            if not labelmaps_json_path.exists():
                raise FileNotFoundError(
                    '`target_type` includes "multi_frame_labels" but '
                    "'labelmaps.json' was not found in root of dataset path:\n"
                    f"{labelmaps_json_path}"
                )
            with labelmaps_json_path.open("r") as fp:
                labelmaps = json.load(fp)
                group = self.training_replicate_metadata.biosound_group
                unit = self.training_replicate_metadata.unit
                id_ = self.training_replicate_metadata.id
            if id_ is not None:
                if group == "Mouse-Pup-Call":
                    self.labelmap = labelmaps[group][unit]["all"]
                else:
                    self.labelmap = labelmaps[group][unit][id_]
            else:
                if group == "Human-Speech":
                    self.labelmap = labelmaps[group][unit]["all"]
                else:
                    raise ValueError(
                        "Unable to determine labelmap to use for "
                        f"group '{group}', unit '{unit}', and id '{id}'. "
                        "Please check that splits_json path is correct."
                    )
        elif target_type == ("binary_frame_labels",):
            self.labelmap = {"no segment": 0, "segment": 1}
        elif target_type == ("boundary_frame_labels",):
            self.labelmap = {"no boundary": 0, "boundary": 1}

        self.split = split
        split_df = pd.read_csv(self.splits_metadata.splits_csv_path)
        split_df = split_df[split_df.split == split].copy()
        self.split_df = split_df

        self.frames_paths = self.split_df[FRAMES_PATH_COL_NAME].values
        self.target_paths = {}
        if "multi_frame_labels" in self.target_type:
            self.target_paths["multi_frame_labels"] = self.split_df[
                MULTI_FRAME_LABELS_PATH_COL_NAME
            ].values
        if "binary_frame_labels" in self.target_type:
            self.target_paths["binary_frame_labels"] = self.split_df[
                BINARY_FRAME_LABELS_PATH_COL_NAME
            ].values
        if "boundary_frame_labels" in self.target_type:
            self.target_paths["boundary_frame_labels"] = self.split_df[
                BOUNDARY_FRAME_LABELS_PATH_COL_NAME
            ].values

        self.window_size = window_size
        self.stride = stride

        # we need all these vectors for getting batches of windows during training
        # for other splits, we use these to determine the duration of the dataset
        self.sample_ids = np.load(
            getattr(self.splits_metadata.sample_id_vector_paths, split)
        )
        self.inds_in_sample = np.load(
            getattr(self.splits_metadata.inds_in_sample_vector_paths, split)
        )
        self.window_inds = (
            datapipes.frame_classification.train_datapipe.get_window_inds(
                self.sample_ids.shape[-1], window_size, stride
            )
        )

        if item_transform is None:
            if standardize_frames and frames_standardizer is None:
                from ..transforms import FramesStandardizer

                frames_standardizer = (
                    FramesStandardizer.fit_inputs_targets_csv_path(
                        self.splits_metadata.splits_csv_path, self.dataset_path
                    )
                )
            if split == "train":
                self.item_transform = TrainItemTransform(
                    frames_standardizer=frames_standardizer
                )
            elif split in ("val", "test", "predict"):
                self.item_transform = InferItemTransform(
                    window_size,
                    frames_standardizer,
                    frames_padval,
                    frame_labels_padval,
                    return_padding_mask,
                )
        else:
            if not callable(item_transform):
                raise ValueError(
                    "`item_transform` should be `callable` "
                    f"but value for `item_transform` was not: {item_transform}"
                )
            self.item_transform = item_transform

        self.return_frames_path = return_frames_path

    @property
    def shape(self):
        tmp_x_ind = 0
        tmp_item = self.__getitem__(tmp_x_ind)
        input_shape = tmp_item["frames"].shape
        if self.split == "train" and len(input_shape) == 3:
            return input_shape
        elif (
            self.split in ("val", "test", "predict") and len(input_shape) == 4
        ):
            # discard windows dimension from shape --
            # it's sample dependent and not what we want
            return input_shape[1:]

    @property
    def duration(self):
        return self.sample_ids.shape[-1] * self.frame_dur

    def __len__(self):
        """number of batches"""
        if self.split == "train":
            return len(self.window_inds)
        else:
            return len(np.unique(self.sample_ids))

    def _getitem_train(self, idx):
        window_idx = self.window_inds[idx]
        sample_ids = self.sample_ids[
            window_idx : window_idx + self.window_size  # noqa: E203
        ]
        uniq_sample_ids = np.unique(sample_ids)
        item = {}
        if len(uniq_sample_ids) == 1:
            # repeat myself to avoid running a loop on one item
            sample_id = uniq_sample_ids[0]
            frames_path = self.dataset_path / self.frames_paths[sample_id]
            spect_dict = common.files.spect.load(frames_path)
            item["frames"] = spect_dict[common.constants.SPECT_KEY]
            for target_type in self.target_type:
                item[target_type] = np.load(
                    self.dataset_path
                    / self.target_paths[target_type][sample_id]
                )

        elif len(uniq_sample_ids) > 1:
            item["frames"] = []
            for target_type in self.target_type:
                # do this to append instead of using defaultdict
                # so that when we do `'target_type' in item` we don't get empty list
                item[target_type] = []
            for sample_id in sorted(uniq_sample_ids):
                frames_path = self.dataset_path / self.frames_paths[sample_id]
                spect_dict = common.files.spect.load(frames_path)
                item["frames"].append(spect_dict[common.constants.SPECT_KEY])
                for target_type in self.target_type:
                    item[target_type].append(
                        np.load(
                            self.dataset_path
                            / self.target_paths[target_type][sample_id]
                        )
                    )

            item["frames"] = np.concatenate(item["frames"], axis=1)
            for target_type in self.target_type:
                item[target_type] = np.concatenate(item[target_type])
        else:
            raise ValueError(
                f"Unexpected number of ``uniq_sample_ids``: {uniq_sample_ids}"
            )

        ind_in_sample = self.inds_in_sample[window_idx]
        item["frames"] = item["frames"][
            ...,
            ind_in_sample : ind_in_sample + self.window_size,  # noqa: E203
        ]
        for target_type in self.target_type:
            item[target_type] = item[target_type][
                ind_in_sample : ind_in_sample + self.window_size  # noqa: E203
            ]
        item = self.item_transform(**item)
        return item

    def _getitem_infer(self, idx):
        item = {}
        frames_path = self.dataset_path / self.frames_paths[idx]
        if self.return_frames_path:
            item["frames_path"] = frames_path
        spect_dict = common.files.spect.load(frames_path)
        item["frames"] = spect_dict[common.constants.SPECT_KEY]
        if self.target_type != "None":  # target_type can be None for predict
            for target_type in self.target_type:
                item[target_type] = np.load(
                    self.dataset_path / self.target_paths[target_type][idx]
                )
        item = self.item_transform(**item)
        return item

    def __getitem__(self, idx):
        if self.split == "train":
            item = self._getitem_train(idx)
        elif self.split in ("val", "test", "predict"):
            item = self._getitem_infer(idx)
        return item



import copy

import pytest
import torch

import vak.models

from .test_factory import ConvEncoderUMAPDefinition



class TestParametricUMAPModel:

    MODEL_DEFINITION_MAP = {
        'ConvEncoderUMAP': ConvEncoderUMAPDefinition,
    }

    @pytest.mark.parametrize(
        'model_name',
        [
            'ConvEncoderUMAP',
        ]
    )
    def test_load_state_dict_from_path(self,
                                       model_name,
                                       specific_config_toml_path,
                                       device
                                       ):
        """Smoke test that makes sure ``load_state_dict_from_path`` runs without failure.

        We use actual model definitions here so we can test with real checkpoints.
        """
        definition = self.MODEL_DEFINITION_MAP[model_name]
        train_toml_path = specific_config_toml_path('train', model_name, audio_format='cbin', annot_format='notmat')
        train_cfg = vak.config.Config.from_toml_path(train_toml_path)

        # stuff we need just to be able to instantiate network
        train_dataset = vak.datapipes.parametric_umap.Datapipe.from_dataset_path(
            dataset_path=train_cfg.train.dataset.path,
            split="train",
        )

        # network is the one thing that has required args
        # and we also need to use its config from the toml file
        cfg = vak.config.Config.from_toml_path(train_toml_path)
        model_config = cfg.train.model.asdict()
        network = {
            'encoder': definition.network['encoder'](
                input_shape=train_dataset.shape,
                **model_config['network']['encoder']
                )
        }
        model_factory = vak.models.factory.ModelFactory(
            definition,
            vak.models.ParametricUMAPModel,
        )
        model = model_factory.from_instances(network=network)
        model.to(device)
        eval_toml_path = specific_config_toml_path('eval', model_name, audio_format='cbin', annot_format='notmat')
        eval_cfg = vak.config.Config.from_toml_path(eval_toml_path)
        checkpoint_path = eval_cfg.eval.checkpoint_path

        # ---- actually test method
        sd_before = copy.deepcopy(model.state_dict())
        sd_before = {
            k: v.to(device) for k, v in sd_before.items()
        }
        ckpt = torch.load(checkpoint_path)
        sd_to_be_loaded = ckpt['state_dict']
        sd_to_be_loaded = {
            k: v.to(device) for k, v in sd_to_be_loaded.items()
        }

        model.load_state_dict_from_path(checkpoint_path)

        assert not all([
            torch.all(torch.eq(val, before_val))
            for val, before_val in zip(model.state_dict().values(), sd_before.values())]
        )
        assert all([
            torch.all(torch.eq(val, before_val))
            for val, before_val in zip(model.state_dict().values(), sd_to_be_loaded.values())]
        )



"""High-level function that trains models."""

from __future__ import annotations

import logging
import pathlib

from .. import models
from ..common import validators
from .frame_classification import train_frame_classification_model
from .parametric_umap import train_parametric_umap_model

logger = logging.getLogger(__name__)


def train(
    model_config: dict,
    dataset_config: dict,
    trainer_config: dict,
    batch_size: int,
    num_epochs: int,
    num_workers: int,
    checkpoint_path: str | pathlib.Path | None = None,
    frames_standardizer_path: str | pathlib.Path | None = None,
    results_path: str | pathlib.Path | None = None,
    standardize_frames: bool = True,
    shuffle: bool = True,
    val_step: int | None = None,
    ckpt_step: int | None = None,
    patience: int | None = None,
    subset: str | None = None,
):
    """Train a model and save results.

    Saves checkpoint files for model,
    label map, and spectrogram scaler.
    These are saved in ``results_path``.

    Parameters
    ----------
    model_config : dict
        Model configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.ModelConfig.asdict`.
    dataset_config: dict
        Dataset configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.DatasetConfig.asdict`.
    trainer_config: dict
        Configuration for :class:`lightning.pytorch.Trainer` in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.TrainerConfig.asdict`.
    window_size : int
        size of windows taken from spectrograms, in number of time bins,
        shown to neural networks
    batch_size : int
        number of samples per batch presented to models during training.
    num_epochs : int
        number of training epochs. One epoch = one iteration through the entire
        training set.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader.
    checkpoint_path : str, pathlib.Path
        Path to a checkpoint file,
        e.g., one generated by a previous run of ``vak.core.train``.
        If specified, this checkpoint will be loaded into model.
        Used when continuing training.
        Default is None, in which case a new model is initialized.
    frames_standardizer_path : str, pathlib.Path
        path to a ``:class:`vak.transforms.FramesStandardizer``` used to standardize (normalize) frames,
        e.g., one generated by a previous run of ``vak.core.train``.
        Used when continuing training, for example on the same dataset.
        Default is None.
    results_path : str, pathlib.Path
        Directory where results will be saved.
    spect_key : str
        key for accessing spectrogram in files. Default is 's'.
    timebins_key : str
        key for accessing vector of time bins in files. Default is 't'.
    device : str
        Device on which to work with model + data.
        Default is None. If None, then a device will be selected with vak.split.get_default.
        That function defaults to 'cuda' if torch.cuda.is_available is True.
    shuffle: bool
        if True, shuffle training data before each epoch. Default is True.
    standardize_frames : bool
        if True, use :class:`vak.transforms.FramesStandardizer` to standardize the frames.
        Normalization is done by subtracting off the mean for each row
        of the training set and then dividing by the std for that frequency bin.
        This same normalization is then applied to validation + test data.
    source_ids : numpy.ndarray
        Parameter for TrainDatapipe. Represents the 'id' of any spectrogram,
        i.e., the index into spect_paths that will let us load it.
        Default is None.
    source_inds : numpy.ndarray
        Parameter for TrainDatapipe. Same length as source_ids
        but values represent indices within each spectrogram.
        Default is None.
    window_inds : numpy.ndarray
        Parameter for TrainDatapipe.
        Indices of each window in the dataset. The value at x[0]
        represents the start index of the first window; using that
        value, we can index into source_ids to get the path
        of the spectrogram file to load, and we can index into
        source_inds to index into the spectrogram itself
        and get the window.
        Default is None.
    val_step : int
        Step on which to estimate accuracy using validation set.
        If val_step is n, then validation is carried out every time
        the global step / n is a whole number, i.e., when val_step modulo the global step is 0.
        Default is None, in which case no validation is done.
    ckpt_step : int
        Step on which to save to checkpoint file.
        If ckpt_step is n, then a checkpoint is saved every time
        the global step / n is a whole number, i.e., when ckpt_step modulo the global step is 0.
        Default is None, in which case checkpoint is only saved at the last epoch.
    patience : int
        number of validation steps to wait without performance on the
        validation set improving before stopping the training.
        Default is None, in which case training only stops after the specified number of epochs.
    split : str
        Name of split from dataset found at ``dataset_path`` to use
        when training model. Default is 'train'. This parameter is used by
        `vak.learncurve.learncurve` to specify specific subsets of the
        training set to use when training models for a learning curve.
    """
    for path, path_name in zip(
        (checkpoint_path, frames_standardizer_path),
        ("checkpoint_path", "frames_standardizer_path"),
    ):
        if path is not None:
            if not validators.is_a_file(path):
                raise FileNotFoundError(
                    f"value for ``{path_name}`` not recognized as a file: {path}"
                )

    dataset_path = pathlib.Path(dataset_config["path"])
    if not dataset_path.exists() or not dataset_path.is_dir():
        raise NotADirectoryError(
            f"`dataset_path` not found or not recognized as a directory: {dataset_path}"
        )

    model_name = model_config["name"]
    try:
        model_family = models.registry.MODEL_FAMILY_FROM_NAME[model_name]
    except KeyError as e:
        raise ValueError(
            f"No model family found for the model name specified: {model_name}"
        ) from e
    if model_family == "FrameClassificationModel":
        train_frame_classification_model(
            model_config=model_config,
            dataset_config=dataset_config,
            trainer_config=trainer_config,
            batch_size=batch_size,
            num_epochs=num_epochs,
            num_workers=num_workers,
            checkpoint_path=checkpoint_path,
            frames_standardizer_path=frames_standardizer_path,
            results_path=results_path,
            standardize_frames=standardize_frames,
            shuffle=shuffle,
            val_step=val_step,
            ckpt_step=ckpt_step,
            patience=patience,
            subset=subset,
        )
    elif model_family == "ParametricUMAPModel":
        train_parametric_umap_model(
            model_config=model_config,
            dataset_config=dataset_config,
            trainer_config=trainer_config,
            batch_size=batch_size,
            num_epochs=num_epochs,
            num_workers=num_workers,
            checkpoint_path=checkpoint_path,
            results_path=results_path,
            shuffle=shuffle,
            val_step=val_step,
            ckpt_step=ckpt_step,
            subset=subset,
        )
    else:
        raise ValueError(f"Model family not recognized: {model_family}")



"""Fixtures made with results from generated test data,
in ./tests/data_for_tests/generated/results"""
from collections import defaultdict

from .model import MODELS
from .test_data import GENERATED_RESULTS_DATA_ROOT


GENERATED_LEARNCURVE_RESULTS_ROOT = GENERATED_RESULTS_DATA_ROOT / 'learncurve'

SOURCE_ANNOT_PAIRS = [
    'audio_cbin_annot_notmat'
]

GENERATED_LEARNCURVE_RESULTS_BY_MODEL = defaultdict(list)

for source_annot_pair in SOURCE_ANNOT_PAIRS:
    source_annot_root = GENERATED_LEARNCURVE_RESULTS_ROOT / source_annot_pair
    for model in MODELS:
        model_root = source_annot_root / model
        results_paths = sorted(model_root.glob('results_*'))
        GENERATED_LEARNCURVE_RESULTS_BY_MODEL[model].extend(results_paths)



from __future__ import annotations

import json
import logging
import pathlib
import warnings

import crowsetta

from ... import datapipes
from ...common import labels
from ...common.converters import expanded_user_path, labelset_to_set
from ...common.logging import config_logging_for_cli, log_version
from ...common.timenow import get_timenow_as_str
from .. import dataset_df_helper, split
from ..unit_dataset import prep_unit_dataset
from . import dataset_arrays

logger = logging.getLogger(__name__)


def prep_parametric_umap_dataset(
    data_dir: str | pathlib.Path,
    purpose: str,
    output_dir: str | pathlib.Path | None = None,
    audio_format: str | None = None,
    spect_params: dict | None = None,
    annot_format: str | None = None,
    annot_file: str | pathlib.Path | None = None,
    labelset: set | None = None,
    context_s: float = 0.015,
    train_dur: int | None = None,
    val_dur: int | None = None,
    test_dur: int | None = None,
    train_set_durs: list[float] | None = None,
    num_replicates: int | None = None,
    spect_key: str = "s",
    timebins_key: str = "t",
):
    """Prepare datasets for neural network models
    that perform a dimensionality reduction task.

    For general information on dataset preparation,
    see the docstring for :func:`vak.prep.prep`.

    Parameters
    ----------
    data_dir : str, Path
        Path to directory with files from which to make dataset.
    purpose : str
        Purpose of the dataset.
        One of {'train', 'eval', 'predict', 'learncurve'}.
        These correspond to commands of the vak command-line interface.
    output_dir : str
        Path to location where data sets should be saved.
        Default is ``None``, in which case it defaults to ``data_dir``.
    audio_format : str
        Format of audio files. One of {'wav', 'cbin'}.
        Default is ``None``, but either ``audio_format`` or ``spect_format``
        must be specified.
    spect_params : dict, vak.config.SpectParams
        Parameters for creating spectrograms. Default is ``None``.
    annot_format : str
        Format of annotations. Any format that can be used with the
        :module:`crowsetta` library is valid. Default is ``None``.
    labelset : str, list, set
        Set of unique labels for vocalizations. Strings or integers.
        Default is ``None``. If not ``None``, then files will be skipped
        where the associated annotation
        contains labels not found in ``labelset``.
        ``labelset`` is converted to a Python ``set`` using
        :func:`vak.converters.labelset_to_set`.
        See help for that function for details on how to specify ``labelset``.
    train_dur : float
        Total duration of training set, in seconds.
        When creating a learning curve,
        training subsets of shorter duration
        will be drawn from this set. Default is None.
    val_dur : float
        Total duration of validation set, in seconds.
        Default is None.
    test_dur : float
        Total duration of test set, in seconds.
        Default is None.
    train_set_durs : list
        of int, durations in seconds of subsets taken from training data
        to create a learning curve, e.g. [5, 10, 15, 20].
    num_replicates : int
        number of times to replicate training for each training set duration
        to better estimate metrics for a training set of that size.
        Each replicate uses a different randomly drawn subset of the training
        data (but of the same duration).
    spect_key : str
        key for accessing spectrogram in files. Default is 's'.
    timebins_key : str
        key for accessing vector of time bins in files. Default is 't'.

    Returns
    -------
    dataset_df : pandas.DataFrame
        That represents a dataset.
    dataset_path : pathlib.Path
        Path to csv saved from ``dataset_df``.
    """
    from .. import constants  # avoid circular import

    # pre-conditions ---------------------------------------------------------------------------------------------------
    if purpose not in constants.VALID_PURPOSES:
        raise ValueError(
            f"purpose must be one of: {constants.VALID_PURPOSES}\n"
            f"Value for purpose was: {purpose}"
        )

    if labelset is not None:
        labelset = labelset_to_set(labelset)

    data_dir = expanded_user_path(data_dir)
    if not data_dir.is_dir():
        raise NotADirectoryError(
            f"Path specified for ``data_dir`` not found: {data_dir}"
        )

    if output_dir:
        output_dir = expanded_user_path(output_dir)
    else:
        output_dir = data_dir

    if not output_dir.is_dir():
        raise NotADirectoryError(
            f"Path specified for ``output_dir`` not found: {output_dir}"
        )

    if annot_file is not None:
        annot_file = expanded_user_path(annot_file)
        if not annot_file.exists():
            raise FileNotFoundError(
                f"Path specified for ``annot_file`` not found: {annot_file}"
            )

    if purpose == "predict":
        if labelset is not None:
            warnings.warn(
                "The ``purpose`` argument was set to 'predict`, but a ``labelset`` was provided."
                "This would cause an error because the ``prep_spectrogram_dataset`` section will attempt to "
                "check whether the files in the ``data_dir`` have labels in "
                "``labelset``, even though those files don't have annotation.\n"
                "Setting ``labelset`` to None."
            )
            labelset = None
    else:  # if purpose is not predict
        if labelset is None:
            raise ValueError(
                f"The ``purpose`` argument was set to '{purpose}', but no ``labelset`` was provided."
                "This will cause an error when trying to split the dataset, "
                "e.g. into training and test splits, "
                "or a silent error, e.g. when calculating metrics with an evaluation set. "
                "Please specify a ``labelset`` when calling ``vak.prep.frame_classification.prep`` "
                f"with ``purpose='{purpose}'."
            )

    logger.info(f"Purpose for frame classification dataset: {purpose}")
    # ---- set up directory that will contain dataset, and csv file name -----------------------------------------------
    data_dir_name = data_dir.name
    timenow = get_timenow_as_str()
    dataset_path = (
        output_dir
        / f"{data_dir_name}-vak-dimensionality-reduction-dataset-generated-{timenow}"
    )
    dataset_path.mkdir()

    if annot_file and annot_format == "birdsong-recognition-dataset":
        # we do this normalization / canonicalization after we make dataset_path
        # so that we can put the new annot_file inside of dataset_path, instead of
        # making new files elsewhere on a user's system
        logger.info(
            "The ``annot_format`` argument was set to 'birdsong-recognition-format'; "
            "this format requires the audio files for their sampling rate "
            "to convert onset and offset times of birdsong syllables to seconds."
            "Converting this format to 'generic-seq' now with the times in seconds, "
            "so that the dataset prepared by vak will not require the audio files."
        )
        birdsongrec = crowsetta.formats.seq.BirdsongRec.from_file(annot_file)
        annots = birdsongrec.to_annot()
        # note we point `annot_file` at a new file we're about to make
        annot_file = (
            dataset_path / f"{annot_file.stem}.converted-to-generic-seq.csv"
        )
        # and we remake Annotations here so that annot_path points to this new file, not the birdsong-rec Annotation.xml
        annots = [
            crowsetta.Annotation(
                seq=annot.seq,
                annot_path=annot_file,
                notated_path=annot.notated_path,
            )
            for annot in annots
        ]
        generic_seq = crowsetta.formats.seq.GenericSeq(annots=annots)
        generic_seq.to_file(annot_file)
        # and we now change `annot_format` as well. Both these will get passed to io.prep_spectrogram_dataset
        annot_format = "generic-seq"

    # NOTE we set up logging here (instead of cli) so the prep log is included in the dataset
    config_logging_for_cli(
        log_dst=dataset_path, log_stem="prep", level="INFO", force=True
    )
    log_version(logger)

    dataset_csv_path = dataset_df_helper.get_dataset_csv_path(
        dataset_path, data_dir_name, timenow
    )
    logger.info(f"Will prepare dataset as directory: {dataset_path}")

    # ---- actually make the dataset -----------------------------------------------------------------------------------
    dataset_df, shape = prep_unit_dataset(
        audio_format=audio_format,
        output_dir=dataset_path,
        spect_params=spect_params,
        data_dir=data_dir,
        annot_format=annot_format,
        annot_file=annot_file,
        labelset=labelset,
        context_s=context_s,
    )

    if dataset_df.empty:
        raise ValueError(
            "Calling `vak.prep.unit_dataset.prep_unit_dataset` "
            "with arguments passed to `vak.core.prep.prep_dimensionality_reduction_dataset` "
            "returned an empty dataframe.\n"
            "Please double-check arguments to `vak.core.prep` function."
        )

    # save before (possibly) splitting, just in case duration args are not valid
    # (we can't know until we make dataset)
    dataset_df.to_csv(dataset_csv_path)

    # ---- (possibly) split into train / val / test sets ---------------------------------------------
    # catch case where user specified duration for just training set, raise a helpful error instead of failing silently
    if (purpose == "train" or purpose == "learncurve") and (
        (train_dur is not None and train_dur > 0)
        and (val_dur is None or val_dur == 0)
        and (test_dur is None or val_dur == 0)
    ):
        raise ValueError(
            "A duration specified for just training set, but prep function does not currently support creating a "
            "single split of a specified duration. Either remove the train_dur option from the prep section and "
            "rerun, in which case all data will be included in the training set, or specify values greater than "
            "zero for test_dur (and val_dur, if a validation set will be used)"
        )

    if all(
        [dur is None for dur in (train_dur, val_dur, test_dur)]
    ) or purpose in (
        "eval",
        "predict",
    ):
        # then we're not going to split
        logger.info("Will not split dataset.")
        do_split = False
    else:
        if val_dur is not None and train_dur is None and test_dur is None:
            raise ValueError(
                "cannot specify only val_dur, unclear how to split dataset into training and test sets"
            )
        else:
            logger.info("Will split dataset.")
            do_split = True

    if do_split:
        dataset_df = split.unit_dataframe(
            dataset_df,
            dataset_path,
            labelset=labelset,
            train_dur=train_dur,
            val_dur=val_dur,
            test_dur=test_dur,
        )

    elif (
        do_split is False
    ):  # add a split column, but assign everything to the same 'split'
        # ideally we would just say split=purpose in call to add_split_col, but
        # we have to special case, because "eval" looks for a 'test' split (not an "eval" split)
        if purpose == "eval":
            split_name = (
                "test"  # 'split_name' to avoid name clash with split package
            )
        elif purpose == "predict":
            split_name = "predict"

        dataset_df = dataset_df_helper.add_split_col(
            dataset_df, split=split_name
        )

    # ---- create and save labelmap ------------------------------------------------------------------------------------
    # we do this before creating array files since we need to load the labelmap to make frame label vectors
    if purpose != "predict":
        # TODO: add option to generate predict using existing dataset, so we can get labelmap from it
        labelmap = labels.to_map(labelset, map_background=False)
        logger.info(
            f"Number of classes in labelmap: {len(labelmap)}",
        )
        # save labelmap in case we need it later
        with (dataset_path / "labelmap.json").open("w") as fp:
            json.dump(labelmap, fp)
    else:
        labelmap = None

    # ---- make arrays that represent final dataset --------------------------------------------------------------------
    dataset_arrays.move_files_into_split_subdirs(
        dataset_df,
        dataset_path,
        purpose,
    )
    #
    # ---- if purpose is learncurve, additionally prep splits for that -----------------------------------------------
    # if purpose == 'learncurve':
    #     dataset_df = make_learncurve_splits_from_dataset_df(
    #         dataset_df,
    #         train_set_durs,
    #         num_replicates,
    #         dataset_path,
    #         labelmap,
    #         audio_format,
    #         spect_key,
    #         timebins_key,
    #     )

    # ---- save csv file that captures provenance of source data -------------------------------------------------------
    logger.info(f"Saving dataset csv file: {dataset_csv_path}")
    dataset_df.to_csv(
        dataset_csv_path, index=False
    )  # index is False to avoid having "Unnamed: 0" column when loading

    # ---- save metadata -----------------------------------------------------------------------------------------------
    metadata = datapipes.parametric_umap.Metadata(
        dataset_csv_filename=str(dataset_csv_path.name),
        audio_format=audio_format,
        shape=shape,
    )
    metadata.to_json(dataset_path)

    return dataset_df, dataset_path



"""tests for vak.cli.learncurve module"""
from unittest import mock

import pytest

import vak.config
import vak.common.constants
import vak.cli.learncurve

from . import cli_asserts


def test_learncurve(specific_config_toml_path, tmp_path, trainer_table):
    root_results_dir = tmp_path.joinpath("test_learncurve_root_results_dir")
    root_results_dir.mkdir()

    keys_to_change = [
        {
            "table": "learncurve",
            "key": "root_results_dir",
            "value": str(root_results_dir),
        },
        {"table": "learncurve", "key": "trainer", "value": trainer_table},
    ]

    toml_path = specific_config_toml_path(
        config_type="learncurve",
        model="TweetyNet",
        audio_format="cbin",
        annot_format="notmat",
        keys_to_change=keys_to_change,
    )

    with mock.patch('vak.learncurve.learning_curve', autospec=True) as mock_core_learning_curve:
        vak.cli.learncurve.learning_curve(toml_path)
        assert mock_core_learning_curve.called

    results_path = sorted(root_results_dir.glob(f"{vak.common.constants.RESULTS_DIR_PREFIX}*"))
    assert len(results_path) == 1
    results_path = results_path[0]

    assert cli_asserts.toml_config_file_copied_to_results_path(results_path, toml_path)
    assert cli_asserts.log_file_created(command="learncurve", output_path=results_path)
    assert cli_asserts.log_file_contains_version(command="learncurve", output_path=results_path)


def test_learning_curve_dataset_none_raises(
        specific_config_toml_path, tmp_path,
):
    """Test that cli.learncurve.learning_curve
    raises ValueError when dataset is None
    (presumably because `vak prep` was not run yet)
    """
    root_results_dir = tmp_path.joinpath("test_learncurve_root_results_dir")
    root_results_dir.mkdir()

    keys_to_change = [
        {
            "table": "learncurve",
            "key": "root_results_dir",
            "value": str(root_results_dir),
        },
        {
            "table": "learncurve",
            "key": "dataset",
            "value": "DELETE-KEY"},
    ]

    toml_path = specific_config_toml_path(
        config_type="learncurve",
        model="TweetyNet",
        audio_format="cbin",
        annot_format="notmat",
        spect_format=None,
        keys_to_change=keys_to_change,
    )

    with pytest.raises(KeyError):
        vak.cli.learncurve.learning_curve(toml_path)



"""sub-package that parses config.toml files and returns config object"""

from . import (
    config,
    dataset,
    eval,
    learncurve,
    load,
    model,
    predict,
    prep,
    spect_params,
    train,
    trainer,
    validators,
)
from .config import Config
from .dataset import DatasetConfig
from .eval import EvalConfig
from .learncurve import LearncurveConfig
from .model import ModelConfig
from .predict import PredictConfig
from .prep import PrepConfig
from .spect_params import SpectParamsConfig
from .train import TrainConfig
from .trainer import TrainerConfig

__all__ = [
    "config",
    "dataset",
    "eval",
    "learncurve",
    "model",
    "load",
    "predict",
    "prep",
    "spect_params",
    "train",
    "trainer",
    "validators",
    "Config",
    "DatasetConfig",
    "EvalConfig",
    "LearncurveConfig",
    "ModelConfig",
    "PredictConfig",
    "PrepConfig",
    "SpectParamsConfig",
    "TrainConfig",
    "TrainerConfig",
]



"""Evaluate a trained model with dataset specified in config.toml file."""

from __future__ import annotations

import logging
import pathlib

from .. import config
from .. import eval as eval_module
from ..common.logging import config_logging_for_cli, log_version

logger = logging.getLogger(__name__)


def eval(toml_path: str | pathlib.Path) -> None:
    """Evaluate a trained model with dataset specified in config.toml file.

    Function called by command-line interface.

    Parameters
    ----------
    toml_path : str, Path
        path to a configuration file in TOML format.

    Returns
    -------
    None
    """
    toml_path = pathlib.Path(toml_path)
    cfg = config.Config.from_toml_path(toml_path)

    if cfg.eval is None:
        raise ValueError(
            f"eval called with a config.toml file that does not have a EVAL section: {toml_path}"
        )

    # ---- set up logging ---------------------------------------------------------------------------------------------
    config_logging_for_cli(
        log_dst=cfg.eval.output_dir, log_stem="eval", level="INFO", force=True
    )
    log_version(logger)

    logger.info("Logging results to {}".format(cfg.eval.output_dir))

    if cfg.eval.dataset.path is None:
        raise ValueError(
            "No value is specified for 'dataset_path' in this .toml config file."
            f"To generate a .csv file that represents the dataset, "
            f"please run the following command:\n'vak prep {toml_path}'"
        )

    eval_module.eval(
        model_config=cfg.eval.model.asdict(),
        dataset_config=cfg.eval.dataset.asdict(),
        trainer_config=cfg.eval.trainer.asdict(),
        checkpoint_path=cfg.eval.checkpoint_path,
        labelmap_path=cfg.eval.labelmap_path,
        output_dir=cfg.eval.output_dir,
        num_workers=cfg.eval.num_workers,
        batch_size=cfg.eval.batch_size,
        frames_standardizer_path=cfg.eval.frames_standardizer_path,
        post_tfm_kwargs=cfg.eval.post_tfm_kwargs,
    )



"""tests for vak.predict module"""
from pathlib import Path

import pandas as pd
import pytest

import vak.config
import vak.common.constants
import vak.predict


# written as separate function so we can re-use in tests/unit/test_cli/test_predict.py
def assert_predict_output_matches_expected(output_dir, annot_csv_filename):
    annot_csv = output_dir.joinpath(annot_csv_filename)
    assert annot_csv.exists()


@pytest.mark.parametrize(
    "model_name, audio_format, spect_format, annot_format, save_net_outputs",
    [
        ("TweetyNet", "cbin", None, "notmat", False),
        ("TweetyNet", "cbin", None, "notmat", True),
    ],
)
def test_predict_with_frame_classification_model(
    model_name,
    audio_format,
    spect_format,
    annot_format,
    save_net_outputs,
        specific_config_toml_path,
    tmp_path,
    trainer_table,
):
    output_dir = tmp_path.joinpath(
        f"test_predict_{audio_format}_{spect_format}_{annot_format}"
    )
    output_dir.mkdir()

    keys_to_change = [
        {"table": "predict", "key": "output_dir", "value": str(output_dir)},
        {"table": "predict", "key": "trainer", "value": trainer_table},
        {"table": "predict", "key": "save_net_outputs", "value": save_net_outputs},
    ]
    toml_path = specific_config_toml_path(
        config_type="predict",
        model=model_name,
        audio_format=audio_format,
        annot_format=annot_format,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)

    vak.predict.frame_classification.predict_with_frame_classification_model(
        model_config=cfg.predict.model.asdict(),
        dataset_config=cfg.predict.dataset.asdict(),
        trainer_config=cfg.predict.trainer.asdict(),
        checkpoint_path=cfg.predict.checkpoint_path,
        labelmap_path=cfg.predict.labelmap_path,
        num_workers=cfg.predict.num_workers,
        timebins_key=cfg.prep.spect_params.timebins_key,
        frames_standardizer_path=cfg.predict.frames_standardizer_path,
        annot_csv_filename=cfg.predict.annot_csv_filename,
        output_dir=cfg.predict.output_dir,
        min_segment_dur=cfg.predict.min_segment_dur,
        majority_vote=cfg.predict.majority_vote,
        save_net_outputs=cfg.predict.save_net_outputs,
    )

    assert_predict_output_matches_expected(output_dir, cfg.predict.annot_csv_filename)
    if save_net_outputs:
        net_outputs = sorted(
            Path(output_dir).glob(f"*{vak.common.constants.NET_OUTPUT_SUFFIX}")
        )

        metadata = vak.datapipes.frame_classification.Metadata.from_dataset_path(cfg.predict.dataset.path)
        dataset_csv_path = cfg.predict.dataset.path / metadata.dataset_csv_filename
        dataset_df = pd.read_csv(dataset_csv_path)

        for spect_path in dataset_df.spect_path.values:
            net_output_spect_path = [
                net_output
                for net_output in net_outputs
                if net_output.name.startswith(Path(spect_path).stem)
            ]
            assert len(net_output_spect_path) == 1


@pytest.mark.parametrize(
    'path_option_to_change',
    [
        {"table": "predict", "key": "checkpoint_path", "value": '/obviously/doesnt/exist/ckpt.pt'},
        {"table": "predict", "key": "labelmap_path", "value": '/obviously/doesnt/exist/labelmap.json'},
        {"table": "predict", "key": "frames_standardizer_path", "value": '/obviously/doesnt/exist/FramesStandardizer'},
    ]
)
def test_predict_with_frame_classification_model_raises_file_not_found(
    path_option_to_change,
        specific_config_toml_path,
    tmp_path,
    trainer_table
):
    """Test that core.eval raises FileNotFoundError
    when `dataset_path` does not exist."""
    output_dir = tmp_path.joinpath(
        f"test_predict_cbin_notmat_invalid_dataset_path"
    )
    output_dir.mkdir()

    keys_to_change = [
        {"table": "predict", "key": "output_dir", "value": str(output_dir)},
        {"table": "predict", "key": "trainer", "value": trainer_table},
        path_option_to_change,
    ]
    toml_path = specific_config_toml_path(
        config_type="predict",
        model="TweetyNet",
        audio_format="cbin",
        annot_format="notmat",
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)

    with pytest.raises(FileNotFoundError):
        vak.predict.frame_classification.predict_with_frame_classification_model(
            model_config=cfg.predict.model.asdict(),
            dataset_config=cfg.predict.dataset.asdict(),
            trainer_config=cfg.predict.trainer.asdict(),
            checkpoint_path=cfg.predict.checkpoint_path,
            labelmap_path=cfg.predict.labelmap_path,
            num_workers=cfg.predict.num_workers,
            timebins_key=cfg.prep.spect_params.timebins_key,
            frames_standardizer_path=cfg.predict.frames_standardizer_path,
            annot_csv_filename=cfg.predict.annot_csv_filename,
            output_dir=cfg.predict.output_dir,
            min_segment_dur=cfg.predict.min_segment_dur,
            majority_vote=cfg.predict.majority_vote,
            save_net_outputs=cfg.predict.save_net_outputs,
        )


@pytest.mark.parametrize(
    'path_option_to_change',
    [
        {"table": "predict", "key": ["dataset", "path"], "value": '/obviously/doesnt/exist/dataset-dir'},
        {"table": "predict", "key": "output_dir", "value": '/obviously/does/not/exist/output'},
    ]
)
def test_predict_with_frame_classification_model_raises_not_a_directory(
    path_option_to_change,
        specific_config_toml_path,
    trainer_table,
    tmp_path,
):
    """Test that core.eval raises NotADirectory
    when ``output_dir`` does not exist
    """
    keys_to_change = [
        path_option_to_change,
        {"table": "predict", "key": "trainer", "value": trainer_table},
    ]

    if path_option_to_change["key"] != "output_dir":
        # need to make sure output_dir *does* exist
        # so we don't detect spurious NotADirectoryError and assume test passes
        output_dir = tmp_path.joinpath(
            f"test_predict_raises_not_a_directory"
        )
        output_dir.mkdir()
        keys_to_change.append(
            {"table": "predict", "key": "output_dir", "value": str(output_dir)}
        )

    toml_path = specific_config_toml_path(
        config_type="predict",
        model="TweetyNet",
        audio_format="cbin",
        annot_format="notmat",
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)

    with pytest.raises(NotADirectoryError):
        vak.predict.frame_classification.predict_with_frame_classification_model(
            model_config=cfg.predict.model.asdict(),
            dataset_config=cfg.predict.dataset.asdict(),
            trainer_config=cfg.predict.trainer.asdict(),
            checkpoint_path=cfg.predict.checkpoint_path,
            labelmap_path=cfg.predict.labelmap_path,
            num_workers=cfg.predict.num_workers,
            timebins_key=cfg.prep.spect_params.timebins_key,
            frames_standardizer_path=cfg.predict.frames_standardizer_path,
            annot_csv_filename=cfg.predict.annot_csv_filename,
            output_dir=cfg.predict.output_dir,
            min_segment_dur=cfg.predict.min_segment_dur,
            majority_vote=cfg.predict.majority_vote,
            save_net_outputs=cfg.predict.save_net_outputs,
        )



from __future__ import annotations

from attrs import asdict, define, field, validators

from .. import common


def is_valid_accelerator(instance, attribute, value):
    """Check if ``accelerator`` is valid"""
    if value == "auto":
        raise ValueError(
            "Using the 'auto' value for the `lightning.pytorch.Trainer` parameter `accelerator` currently "
            "breaks functionality for the command-line interface of `vak`. "
            "Please see this issue: https://github.com/vocalpy/vak/issues/691"
            "If you need to use the 'auto' mode of `lightning.pytorch.Trainer`, please use `vak` directly in a script."
        )
    elif value in ("cpu", "gpu", "tpu", "ipu"):
        return
    else:
        raise ValueError(
            f"Invalid value for 'accelerator' key in 'trainer' table of configuration file: {value}. "
            'Value must be one of: {"cpu", "gpu", "tpu", "ipu"}'
        )


def is_valid_devices(instance, attribute, value):
    """Check if ``devices`` is valid"""
    if not (
        (isinstance(value, int))
        or (
            isinstance(value, list)
            and all([isinstance(el, int) for el in value])
        )
    ):
        raise ValueError(
            "Invalid value for 'devices' key in 'trainer' table of configuration file: {value}"
        )


@define
class TrainerConfig:
    """Class that represents ``trainer`` sub-table
    in a toml configuration file.

    Used to configure :class:`lightning.pytorch.Trainer`.

    Attributes
    ----------
    accelerator : str
        Value for the `accelerator` argument to :class:`lightning.pytorch.Trainer`.
        Default is the return value of :func:`vak.common.accelerator.get_default`.
    devices: int, list of int
        Number of devices (int) or exact device(s) (list of int) to use.

    Notes
    -----
    Using the 'auto' value for the `lightning.pytorch.Trainer` parameter `accelerator` currently
    breaks functionality for the command-line interface of `vak`.
    Please see this issue: https://github.com/vocalpy/vak/issues/691
    If you need to use the 'auto' mode of `lightning.pytorch.Trainer`, please use `vak` directly in a script.

    Likewise, setting a value for the `lightning.pytorch.Trainer` parameter `devices` that is not either 1
    (meaning \"use a single GPU\") or a list with a single number (meaning \"use this exact GPU\") currently
    breaks functionality for the command-line interface of `vak`.
    Please see this issue: https://github.com/vocalpy/vak/issues/691
    If you need to use multiple GPUs, please use `vak` directly in a script.
    """

    accelerator: str = field(
        validator=is_valid_accelerator,
        default=common.accelerator.get_default(),
    )
    devices: int | list[int] = field(
        validator=validators.optional(is_valid_devices),
        # for devices, we need to look at accelerator in post-init to determine default
        default=None,
    )

    def __attrs_post_init__(self):
        # set default self.devices *before* we validate,
        # so that we don't throw error because of the default None
        # that we need to change here depending on the value of self.accelerator
        if self.devices is None:
            if self.accelerator == "cpu":
                # ~"use all available"
                self.devices = 1
            elif self.accelerator in ("gpu", "tpu", "ipu"):
                # we can only use a single device, assume there's only one
                self.devices = [0]

        if self.accelerator in ("gpu", "tpu", "ipu"):
            if not (
                (isinstance(self.devices, int) and self.devices == 1)
                or (
                    isinstance(self.devices, list)
                    and len(self.devices) == 1
                    and all([isinstance(el, int) for el in self.devices])
                )
            ):
                raise ValueError(
                    "Setting a value for the `lightning.pytorch.Trainer` parameter `devices` that is not either 1 "
                    '(meaning "use a single GPU") or a list with a single number '
                    '(meaning "use this exact GPU") currently '
                    "breaks functionality for the command-line interface of `vak`. "
                    "Please see this issue: https://github.com/vocalpy/vak/issues/691"
                    "If you need to use multiple GPUs, please use `vak` directly in a script."
                )
        elif self.accelerator == "cpu":
            if isinstance(self.devices, list):
                raise ValueError(
                    f"Value for `devices` cannot be a list when `accelerator` is `cpu`. Value was: {self.devices}\n"
                    "When `accelerator` is `cpu`, please set `devices` to 1 or -1 (which are equivalent)."
                )
            if self.devices < 1:
                raise ValueError(
                    "When value for 'accelerator' is 'cpu', value for `devices` "
                    f"should be an int > 0, but was: {self.devices}"
                )

    def asdict(self):
        """Convert this :class:`TrainerConfig` instance
        to a :class:`dict` that can be passed
        into functions that take a ``trainer_config`` argument,
        like :func:`vak.train` and :func:`vak.predict`.
        """
        return asdict(self)



"""Function that generates results for a learning curve for frame classification models."""

from __future__ import annotations

import logging
import pathlib

import pandas as pd

from .. import common, datapipes
from ..common.converters import expanded_user_path
from ..eval.frame_classification import eval_frame_classification_model
from ..train.frame_classification import train_frame_classification_model
from .dirname import replicate_dirname, train_dur_dirname

logger = logging.getLogger(__name__)


def learning_curve_for_frame_classification_model(
    model_config: dict,
    dataset_config: dict,
    trainer_config: dict,
    batch_size: int,
    num_epochs: int,
    num_workers: int,
    results_path: str | pathlib.Path,
    post_tfm_kwargs: dict | None = None,
    standardize_frames: bool = True,
    shuffle: bool = True,
    val_step: int | None = None,
    ckpt_step: int | None = None,
    patience: int | None = None,
) -> None:
    """Generate results for a learning curve,
    where model performance is measured as a
    function of dataset duration in seconds.

    Trains a class of model with a range of dataset durations,
    and then evaluates each trained model
    with a test set that is held constant
    (unlike the training sets).
    Results are saved in a new directory within ``root_results_dir``.

    Parameters
    ----------
    model_config : dict
        Model configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.ModelConfig.asdict`.
    dataset_config: dict
        Dataset configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.DatasetConfig.asdict`.
    trainer_config: dict
        Configuration for :class:`lightning.pytorch.Trainer`.
        Can be obtained by calling :meth:`vak.config.TrainerConfig.asdict`.
    dataset_path : str
        path to where dataset was saved as a csv.
    batch_size : int
        number of samples per batch presented to models during training.
    num_epochs : int
        number of training epochs. One epoch = one iteration through the entire
        training set.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader.
    results_path : str, pathlib.Path
        Directory where results will be saved.
    previous_run_path : str, Path
        Path to directory containing dataset .csv files
        that represent subsets of training set, created by
        a previous run of ``vak.core.learncurve.learning_curve``.
        Typically directory will have a name like ``results_{timestamp}``
        and the actual .csv splits will be in sub-directories with names
        corresponding to the training set duration
    post_tfm_kwargs : dict
        Keyword arguments to post-processing transform.
        If None, then no additional clean-up is applied
        when transforming labeled timebins to segments,
        the default behavior. The transform used is
        ``vak.transforms.frame_labels.ToSegmentsWithPostProcessing`.
        Valid keyword argument names are 'majority_vote'
        and 'min_segment_dur', and should be appropriate
        values for those arguments: Boolean for ``majority_vote``,
        a float value for ``min_segment_dur``.
        See the docstring of the transform for more details on
        these arguments and how they work.
    device : str
        Device on which to work with model + data.
        Default is None. If None, then a device will be selected with vak.device.get_default.
        That function defaults to 'cuda' if torch.cuda.is_available is True.
    shuffle: bool
        if True, shuffle training data before each epoch. Default is True.
    standardize_frames : bool
        if True, use :class:`vak.transforms.FramesStandardizer` to standardize the frames.
        Normalization is done by subtracting off the mean for each row
        of the training set and then dividing by the std for that frequency bin.
        This same normalization is then applied to validation + test data.
    val_step : int
        Step on which to estimate accuracy using validation set.
        If val_step is n, then validation is carried out every time
        the global step / n is a whole number, i.e., when val_step modulo the global step is 0.
        Default is None, in which case no validation is done.
    ckpt_step : int
        Step on which to save to checkpoint file.
        If ckpt_step is n, then a checkpoint is saved every time
        the global step / n is a whole number, i.e., when ckpt_step modulo the global step is 0.
        Default is None, in which case checkpoint is only saved at the last epoch.
    patience : int
        number of validation steps to wait without performance on the
        validation set improving before stopping the training.
        Default is None, in which case training only stops after the specified number of epochs.
    """
    # ---------------- pre-conditions ----------------------------------------------------------------------------------
    dataset_path = expanded_user_path(dataset_config["path"])
    if not dataset_path.exists() or not dataset_path.is_dir():
        raise NotADirectoryError(
            f"`dataset_path` not found or not recognized as a directory: {dataset_path}"
        )

    logger.info(
        f"Loading dataset from path: {dataset_path}",
    )
    metadata = datapipes.frame_classification.Metadata.from_dataset_path(
        dataset_path
    )
    dataset_csv_path = dataset_path / metadata.dataset_csv_filename
    dataset_df = pd.read_csv(dataset_csv_path)

    if val_step and not dataset_df["split"].str.contains("val").any():
        raise ValueError(
            f"val_step set to {val_step} but dataset does not contain a validation set; "
            f"please run `vak prep` with a config.toml file that specifies a duration for the validation set."
        )

    # ---- set up directory to save output -----------------------------------------------------------------------------
    results_path = expanded_user_path(results_path)
    if not results_path.is_dir():
        raise NotADirectoryError(
            f"results_path not recognized as a directory: {results_path}"
        )

    logger.info(f"Saving results to: {results_path}")

    frame_dur = metadata.frame_dur
    logger.info(
        f"Duration of a frame in dataset, in seconds: {frame_dur}",
    )

    # ---- get training set subsets ------------------------------------------------------------------------------------
    dataset_df = dataset_df[
        (dataset_df.train_dur.notna()) & (dataset_df.replicate_num.notna())
    ]
    train_durs = sorted(dataset_df["train_dur"].unique())
    replicate_nums = [
        int(replicate_num)
        for replicate_num in sorted(dataset_df["replicate_num"].unique())
    ]
    to_do = []
    for train_dur in train_durs:
        for replicate_num in replicate_nums:
            to_do.append((train_dur, replicate_num))

    # ---- main loop that creates "learning curve" ---------------------------------------------------------------------
    logger.info("Starting training for learning curve.")
    model_name = model_config[
        "name"
    ]  # used below when getting checkpoint path, etc
    for train_dur, replicate_num in to_do:
        logger.info(
            f"Training model with training set of size: {train_dur}s, replicate number {replicate_num}.",
        )
        results_path_this_train_dur = results_path / train_dur_dirname(
            train_dur
        )
        if not results_path_this_train_dur.exists():
            results_path_this_train_dur.mkdir()

        results_path_this_replicate = (
            results_path_this_train_dur / replicate_dirname(replicate_num)
        )
        results_path_this_replicate.mkdir()

        logger.info(
            f"Saving results to: {results_path_this_replicate}",
        )

        # `subset` lets us use correct subset of training set for this duration / replicate
        subset = common.learncurve.get_train_dur_replicate_subset_name(
            train_dur, replicate_num
        )

        train_frame_classification_model(
            model_config,
            dataset_config,
            trainer_config,
            batch_size,
            num_epochs,
            num_workers,
            results_path=results_path_this_replicate,
            standardize_frames=standardize_frames,
            shuffle=shuffle,
            val_step=val_step,
            ckpt_step=ckpt_step,
            patience=patience,
            subset=subset,
        )

        logger.info(f"Evaluating model from replicate {replicate_num} ")
        results_model_root = results_path_this_replicate.joinpath(model_name)
        ckpt_root = results_model_root.joinpath("checkpoints")
        ckpt_paths = sorted(ckpt_root.glob("*.pt"))
        if any(["max-val-acc" in str(ckpt_path) for ckpt_path in ckpt_paths]):
            ckpt_paths = [
                ckpt_path
                for ckpt_path in ckpt_paths
                if "max-val-acc" in str(ckpt_path)
            ]
            if len(ckpt_paths) != 1:
                raise ValueError(
                    f"did not find a single max-val-acc checkpoint path, instead found:\n{ckpt_paths}"
                )
            ckpt_path = ckpt_paths[0]
        else:
            if len(ckpt_paths) != 1:
                raise ValueError(
                    f"did not find a single checkpoint path, instead found:\n{ckpt_paths}"
                )
            ckpt_path = ckpt_paths[0]
        logger.info(f"Using checkpoint: {ckpt_path}")
        labelmap_path = results_path_this_replicate.joinpath("labelmap.json")
        logger.info(f"Using labelmap: {labelmap_path}")
        if standardize_frames:
            frames_standardizer_path = results_path_this_replicate.joinpath(
                "FramesStandardizer"
            )
            logger.info(
                f"Using FramesStandardizer to standardize frames, from path: {frames_standardizer_path}",
            )
        else:
            frames_standardizer_path = None

        eval_frame_classification_model(
            model_config,
            dataset_config,
            trainer_config,
            ckpt_path,
            labelmap_path,
            results_path_this_replicate,
            num_workers,
            frames_standardizer_path,
            post_tfm_kwargs,
        )

    # ---- make a csv for analysis -------------------------------------------------------------------------------------
    # use one of the eval csvs just to get columns, that we use below to re-order
    eval_csv_paths = sorted(results_path.glob("**/eval*.csv"))
    eval_df_0 = pd.read_csv(eval_csv_paths[0])
    eval_columns = eval_df_0.columns.tolist()

    eval_dfs = []
    for train_dur, replicate_num in to_do:
        results_path_this_train_dur = results_path / train_dur_dirname(
            train_dur
        )
        results_path_this_replicate = (
            results_path_this_train_dur / replicate_dirname(replicate_num)
        )
        eval_csv_path = sorted(results_path_this_replicate.glob("eval*.csv"))
        if not len(eval_csv_path) == 1:
            raise ValueError(
                "Did not find exactly one eval results csv file in replicate directory after running learncurve. "
                f"Directory is: {results_path_this_replicate}."
                f"Result of globbing directory for eval csv file: {eval_csv_path}"
            )
        eval_csv_path = eval_csv_path[0]
        eval_df = pd.read_csv(eval_csv_path)
        eval_df["train_dur"] = train_dur
        eval_df["replicate_num"] = replicate_num
        eval_dfs.append(eval_df)

    all_eval_df = pd.concat(eval_dfs)
    all_eval_columns = ["train_dur", "replicate_num", *eval_columns]
    all_eval_df = all_eval_df[all_eval_columns]
    all_eval_df.sort_values(by=["train_dur", "replicate_num"])
    learncurve_csv_path = results_path.joinpath("learning_curve.csv")
    all_eval_df.to_csv(
        learncurve_csv_path, index=False
    )  # index=False to avoid adding "Unnamed: 0" column



"""Functions to parse toml config files."""

from __future__ import annotations

import pathlib

import tomlkit
import tomlkit.exceptions


def _tomlkit_to_popo(d):
    """Convert tomlkit to "popo" (Plain-Old Python Objects)

    From https://github.com/python-poetry/tomlkit/issues/43#issuecomment-660415820

    We need this so we don't get a ``tomlkit.items._ConvertError`` when
    the `from_config_dict` classmethods try to add a class to a ``config_dict``,
    e.g. when :meth:`EvalConfig.from_config_dict` converts the ``spect_params``
    key-value pairs to a :class:`vak.config.SpectParamsConfig` instance
    and then assigns it to the ``spect_params`` key.
    We would get this error if we just return the result of :func:`tomlkit.load`,
    which is a `tomlkit.TOMLDocument` that tries to ensure that everything is valid toml.
    """
    try:
        result = getattr(d, "value")
    except AttributeError:
        result = d

    if isinstance(result, list):
        result = [_tomlkit_to_popo(x) for x in result]
    elif isinstance(result, dict):
        result = {
            _tomlkit_to_popo(key): _tomlkit_to_popo(val)
            for key, val in result.items()
        }
    elif isinstance(result, tomlkit.items.Integer):
        result = int(result)
    elif isinstance(result, tomlkit.items.Float):
        result = float(result)
    elif isinstance(result, tomlkit.items.String):
        result = str(result)
    elif isinstance(result, tomlkit.items.Bool):
        result = bool(result)

    return result


def _load_toml_from_path(toml_path: str | pathlib.Path) -> dict:
    """Load a toml file from a path, and return as a :class:`dict`.

    Notes
    -----
    Helper function to load toml config file,
    factored out to use in other modules when needed.
    Checks if ``toml_path`` exists before opening,
    and tries to give a clear message if an error occurs when loading.

    Note also this function checks that the loaded :class:`dict`
    has a single top-level key ``'vak'``,
    and that it returns the :class:`dict` one level down
    that is accessed with that key.
    This avoids the need to write ``['vak']`` everywhere in
    calling functions.
    However it also means you need to add back that key
    if you are *writing* a toml file.
    """
    toml_path = pathlib.Path(toml_path)
    if not toml_path.is_file():
        raise FileNotFoundError(f".toml config file not found: {toml_path}")

    try:
        with toml_path.open("r") as fp:
            config_dict: dict = tomlkit.load(fp)
    except tomlkit.exceptions.TOMLKitError as e:
        raise Exception(
            f"Error when parsing .toml config file: {toml_path}"
        ) from e

    if "vak" not in config_dict:
        raise ValueError(
            "Toml file does not contain a top-level table named `vak`. "
            "Please see example configuration files here:\n"
            "https://github.com/vocalpy/vak/tree/main/doc/toml"
        )

    # Next line, convert TOMLDocument returned by tomlkit.load to a dict.
    # We need this so we don't get a ``tomlkit.items._ConvertError`` when
    # the `from_config_dict` classmethods try to add a class to a ``config_dict``,
    # e.g. when :meth:`EvalConfig.from_config_dict` converts the ``spect_params``
    # key-value pairs to a :class:`vak.config.SpectParamsConfig` instance
    # and then assigns it to the ``spect_params`` key.
    # We would get this error if we just return the result of :func:`tomlkit.load`,
    # which is a `tomlkit.TOMLDocument` that tries to ensure that everything is valid toml.
    return _tomlkit_to_popo(config_dict)["vak"]



"""A dataclass that represents metadata
associated with a frame classification dataset,
as generated by
:func:`vak.core.prep.frame_classification.prep_frame_classification_dataset`"""

from __future__ import annotations

import json
import pathlib
from typing import ClassVar

import attr


def is_valid_dataset_csv_filename(instance, attribute, value):
    valid = "_prep_" in value and value.endswith(".csv")
    if not valid:
        raise ValueError(
            f"Invalid dataset csv filename: {value}."
            f'Filename should contain the string "_prep_" '
            f"and end with the extension .csv."
            f"Valid filenames are generated by "
            f"vak.core.prep.generate_dataset_csv_filename"
        )


def is_valid_audio_format(instance, attribute, value):
    import vak.common.constants

    if value not in vak.common.constants.VALID_AUDIO_FORMATS:
        raise ValueError(
            f"Not a valid audio format: {value}. Valid audio formats are: {vak.common.constants.VALID_AUDIO_FORMATS}"
        )


def is_valid_spect_format(instance, attribute, value):
    import vak.common.constants

    if value not in vak.common.constants.VALID_SPECT_FORMATS:
        raise ValueError(
            f"Not a valid spectrogram format: {value}. "
            f"Valid spectrogram formats are: {vak.common.constants.VALID_SPECT_FORMATS}"
        )


@attr.define
class Metadata:
    """A dataclass that represents metadata
    associated with a dataset that was
    generated by :func:`vak.core.prep.prep`.

    Attributes
    ----------
    dataset_csv_filename : str
        Name of csv file representing the source files in the dataset.
        Csv file will be located in root of directory representing dataset,
        so only the filename is given.
    input_type : str
        The modality of the input data "frames", either audio signals
        or spectrograms. One of {'audio', 'spect'}.
    frame_dur: float
        Duration of a frame, i.e., a single sample in audio
        or a single timebin in a spectrogram.
    """

    # declare this as a constant to avoid
    # needing to remember this in multiple places, and to use in unit tests
    METADATA_JSON_FILENAME: ClassVar = "metadata.json"

    dataset_csv_filename: str = attr.field(
        converter=str, validator=is_valid_dataset_csv_filename
    )

    input_type: str = attr.field()

    @input_type.validator
    def is_valid_input_type(self, attribute, value):
        if not isinstance(value, str):
            raise TypeError(
                f"{attribute.name} value should be a string but was type {type(value)}"
            )
        from ...prep.constants import INPUT_TYPES

        if value not in INPUT_TYPES:
            raise ValueError(
                f"Value for {attribute.name} is not a valid input type: '{value}'\n"
                f"Valid input types are: {INPUT_TYPES}"
            )

    frame_dur: float = attr.field(converter=float)

    @frame_dur.validator
    def is_valid_frame_dur(self, attribute, value):
        if not isinstance(value, float):
            raise ValueError(f"{attribute.name} should be a float value.")
        if not value > 0.0:
            raise ValueError(f"{attribute.name} should be greater than zero.")

    audio_format: str = attr.field(
        converter=attr.converters.optional(str),
        validator=attr.validators.optional(is_valid_audio_format),
        default=None,
    )

    spect_format: str = attr.field(
        converter=attr.converters.optional(str),
        validator=attr.validators.optional(is_valid_spect_format),
        default=None,
    )

    @classmethod
    def from_path(cls, json_path: str | pathlib.Path) -> Metadata:
        """Load dataset metadata from a json file.

        Class method that returns an instance of
        :class:`~vak.datasets.frame_classification.FrameClassificationDatatsetMetadata`.

        Parameters
        ----------
        json_path : string, pathlib.Path
            Path to a 'metadata.json' file created by
            :func:`vak.core.prep.prep` when generating
            a dataset.

        Returns
        -------
        metadata : vak.datasets.frame_classification.FrameClassificationDatatsetMetadata
            Instance of :class:`~vak.datasets.frame_classification.FrameClassificationDatatsetMetadata`
            with metadata loaded from json file.
        """
        json_path = pathlib.Path(json_path)
        with json_path.open("r") as fp:
            metadata_json = json.load(fp)
        return cls(**metadata_json)

    @classmethod
    def from_dataset_path(cls, dataset_path: str | pathlib.Path):
        dataset_path = pathlib.Path(dataset_path)
        if not dataset_path.exists() or not dataset_path.is_dir():
            raise NotADirectoryError(
                f"`dataset_path` not found or not recognized as a directory: {dataset_path}"
            )

        metadata_json_path = dataset_path / cls.METADATA_JSON_FILENAME
        if not metadata_json_path.exists():
            raise FileNotFoundError(
                f"Metadata file not found: {metadata_json_path}"
            )

        return cls.from_path(metadata_json_path)

    def to_json(self, dataset_path: str | pathlib.Path) -> None:
        """Dump dataset metadata to a json file.

        This method is called by :func:`vak.core.prep.prep`
        after it generates a dataset and then creates an
        instance of :class:`~vak.datasets.frame_classification.FrameClassificationDatatsetMetadata`
        with metadata about that dataset.

        Parameters
        ----------
        dataset_path : string, pathlib.Path
            Path to root of a directory representing a dataset
            generated by :func:`vak.core.prep.prep`.
            where 'metadata.json' file
            should be saved.
        """
        dataset_path = pathlib.Path(dataset_path)
        if not dataset_path.exists() or not dataset_path.is_dir():
            raise NotADirectoryError(
                f"dataset_path not recognized as a directory: {dataset_path}"
            )

        json_dict = attr.asdict(self)
        json_path = dataset_path / self.METADATA_JSON_FILENAME
        with json_path.open("w") as fp:
            json.dump(json_dict, fp, indent=4)



"""Parametric UMAP model, as described in [1]_.

Code adapted from implementation by @elyxlz
https://github.com/elyxlz/umap_pytorch
with changes made by Tim Sainburg:
https://github.com/lmcinnes/umap/issues/580#issuecomment-1368649550.
"""

from __future__ import annotations

import pathlib
from typing import Callable, ClassVar, Type

import lightning
import torch
import torch.utils.data

from .definition import ModelDefinition
from .registry import model_family


@model_family
class ParametricUMAPModel(lightning.LightningModule):
    """Parametric UMAP model, as described in [1]_.

    Notes
    -----
    Code adapted from implementation by @elyxlz
    https://github.com/elyxlz/umap_pytorch
    with changes made by Tim Sainburg:
    https://github.com/lmcinnes/umap/issues/580#issuecomment-1368649550.

    References
    ----------
    .. [1] Sainburg, T., McInnes, L., & Gentner, T. Q. (2021).
       Parametric UMAP embeddings for representation and semisupervised learning.
       Neural Computation, 33(11), 2881-2907.
       https://direct.mit.edu/neco/article/33/11/2881/107068.
    """

    definition: ClassVar[ModelDefinition]

    def __init__(
        self,
        network: dict,
        loss: torch.nn.Module | Callable,
        optimizer: torch.optim.Optimizer,
        metrics: dict[str:Type],
    ):
        super().__init__()
        self.network = torch.nn.ModuleDict(network)
        self.loss = loss
        self.optimizer = optimizer
        self.metrics = metrics

    def configure_optimizers(self):
        return self.optimizer

    def training_step(self, batch, batch_idx):
        (edges_to_exp, edges_from_exp) = batch
        embedding_to = self.network["encoder"](edges_to_exp)
        embedding_from = self.network["encoder"](edges_from_exp)

        if "decoder" in self.network:
            reconstruction = self.network["decoder"](embedding_to)
            before_encoding = edges_to_exp
        else:
            reconstruction = None
            before_encoding = None
        loss_umap, loss_reconstruction, loss = self.loss(
            embedding_to, embedding_from, reconstruction, before_encoding
        )
        self.log("train_umap_loss", loss_umap, on_step=True)
        if loss_reconstruction:
            self.log(
                "train_reconstruction_loss", loss_reconstruction, on_step=True
            )
        # note if there's no ``loss_reconstruction``, then ``loss`` == ``loss_umap``
        self.log("train_loss", loss, on_step=True)
        return loss

    def validation_step(self, batch, batch_idx):
        (edges_to_exp, edges_from_exp) = batch
        embedding_to = self.network["encoder"](edges_to_exp)
        embedding_from = self.network["encoder"](edges_from_exp)

        if "decoder" in self.network is not None:
            reconstruction = self.network["decoder"](embedding_to)
            before_encoding = edges_to_exp
        else:
            reconstruction = None
            before_encoding = None
        loss_umap, loss_reconstruction, loss = self.loss(
            embedding_to, embedding_from, reconstruction, before_encoding
        )
        self.log("val_umap_loss", loss_umap, on_step=True)
        if loss_reconstruction:
            self.log(
                "val_reconstruction_loss", loss_reconstruction, on_step=True
            )
        # note if there's no ``loss_reconstruction``, then ``loss`` == ``loss_umap``
        self.log("val_loss", loss, on_step=True)

    def load_state_dict_from_path(self, ckpt_path):
        """Loads a model from the path to a saved checkpoint.

        Loads the checkpoint and then calls
        ``self.load_state_dict`` with the ``state_dict``
        in that chekcpoint.

        This method allows loading a state dict into an instance.
        It's necessary because `lightning.pytorch.LightningModule.load`` is a
        ``classmethod``, so calling that method will trigger
         ``LightningModule.__init__`` instead of running
        ``vak.models.Model.__init__``.

        Parameters
        ----------
        ckpt_path : str, pathlib.Path
            Path to a checkpoint saved by a model in ``vak``.
            This checkpoint has the same key-value pairs as
            any other checkpoint saved by a
            ``lightning.pytorch.LightningModule``.

        Returns
        -------
        None

        This method modifies the model state by loading the ``state_dict``;
        it does not return anything.
        """
        ckpt = torch.load(ckpt_path)
        self.load_state_dict(ckpt["state_dict"])


class ParametricUMAPDatamodule(lightning.pytorch.LightningDataModule):
    def __init__(
        self,
        dataset,
        batch_size,
        num_workers,
    ):
        super().__init__()
        self.dataset = dataset
        self.batch_size = batch_size
        self.num_workers = num_workers

    def train_dataloader(self) -> torch.utils.data.DataLoader:
        return torch.utils.data.DataLoader(
            dataset=self.dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=True,
        )


class ParametricUMAP:
    def __init__(
        self,
        encoder: torch.nn.Module,
        decoder: torch.nn.Module | None = None,
        n_neighbors: int = 10,
        min_dist: float = 0.1,
        metric: str = "euclidean",
        num_epochs: int = 200,
        lr: float = 1e-3,
        batch_size: int = 64,
        num_workers: int = 16,
        random_state: int | None = None,
    ):
        self.encoder = encoder
        self.decoder = decoder
        self.n_neighbors = n_neighbors
        self.min_dist = min_dist
        self.metric = metric

        self.lr = lr
        self.num_epochs = num_epochs

        self.batch_size = batch_size
        self.num_workers = num_workers
        self.random_state = random_state

        self.model = ParametricUMAPModel(self.encoder, min_dist=self.min_dist)

    def fit(
        self,
        trainer: lightning.pytorch.Trainer,
        dataset_path: str | pathlib.Path,
        transform=None,
    ):
        from vak.datapipes.parametric_umap import ParametricUMAPDataset

        dataset = ParametricUMAPDataset.from_dataset_path(
            dataset_path,
            "train",
            self.n_neighbors,
            self.metric,
            self.random_state,
            self.num_epochs,
            transform,
        )
        trainer.fit(
            model=self.model,
            datamodule=ParametricUMAPDatamodule(
                dataset, self.batch_size, self.num_workers
            ),
        )

    @torch.no_grad()
    def transform(self, X):
        embedding = self.model.encoder(X).detach().cpu().numpy()
        return embedding

    @torch.no_grad()
    def inverse_transform(self, Z):
        return self.model.decoder(Z).detach().cpu().numpy()



"""Helper functions for `vak.prep.dimensionality_reduction` module
that handle array files.
"""

from __future__ import annotations

import logging
import pathlib
import shutil

import pandas as pd

logger = logging.getLogger(__name__)


def move_files_into_split_subdirs(
    dataset_df: pd.DataFrame, dataset_path: pathlib.Path, purpose: str
) -> None:
    """Move npy files in dataset into sub-directories, one for each split in the dataset.

    This is run *after* calling :func:`vak.prep.unit_dataset.prep_unit_dataset`
    to generate ``dataset_df``.

    Parameters
    ----------
    dataset_df : pandas.DataFrame
        A ``pandas.DataFrame`` returned by
        :func:`vak.prep.unit_dataset.prep_unit_dataset`
        with a ``'split'`` column added, as a result of calling
        :func:`vak.prep.split.unit_dataframe` or because it was added "manually"
        by calling :func:`vak.core.prep.prep_helper.add_split_col` (as is done
        for 'predict' when the entire ``DataFrame`` belongs to this
        "split").
    dataset_path : pathlib.Path
        Path to directory that represents dataset.
    purpose: str
        A string indicating what the dataset will be used for.
        One of {'train', 'eval', 'predict', 'learncurve'}.
        Determined by :func:`vak.core.prep.prep`
        using the TOML configuration file.

    Returns
    -------
    None

    The ``DataFrame`` is modified in place
    as the files are moved, so nothing is returned.
    """
    moved_spect_paths = (
        []
    )  # to clean up after moving -- may be empty if we copy all spects (e.g., user generated)
    # ---- copy/move files into split sub-directories inside dataset directory
    # Next line, note we drop any na rows in the split column, since they don't belong to a split anyway
    split_names = sorted(dataset_df.split.dropna().unique())

    for split_name in split_names:
        if split_name == "None":
            # these are files that didn't get assigned to a split
            continue
        split_subdir = dataset_path / split_name
        split_subdir.mkdir()

        split_df = dataset_df[dataset_df.split == split_name].copy()
        split_spect_paths = [
            # this just converts from string to pathlib.Path
            pathlib.Path(spect_path)
            for spect_path in split_df["spect_path"].values
        ]
        is_in_dataset_dir = [
            # if dataset_path is one of the parents of spect_path, we can move; otherwise, we copy
            dataset_path.resolve() in list(spect_path.parents)
            for spect_path in split_spect_paths
        ]
        if all(is_in_dataset_dir):
            move_spects = True
        elif all([not is_in_dir for is_in_dir in is_in_dataset_dir]):
            move_spects = False
        else:
            raise ValueError(
                "Expected to find either all spectrograms were in dataset directory, "
                "or all were in some other directory, but found a mixture. "
                f"Spectrogram paths for split being moved within dataset directory:\n{split_spect_paths}"
            )

        new_spect_paths = []  # to fix DataFrame
        for spect_path in split_spect_paths:
            spect_path = pathlib.Path(spect_path)
            if move_spects:  # because it's within dataset_path already
                new_spect_path = spect_path.rename(
                    split_subdir / spect_path.name
                )
                moved_spect_paths.append(spect_path)
            else:  # copy instead of moving
                new_spect_path = shutil.copy(src=spect_path, dst=split_subdir)

            new_spect_paths.append(
                # rewrite paths relative to dataset directory's root, so dataset is portable
                pathlib.Path(new_spect_path).relative_to(dataset_path)
            )

        # cast to str before rewrite so that type doesn't silently change for some rows
        new_spect_paths = [
            str(new_spect_path) for new_spect_path in new_spect_paths
        ]
        dataset_df.loc[split_df.index, "spect_path"] = new_spect_paths

    # ---- clean up after moving/copying -------------------------------------------------------------------------------
    # remove any directories that we just emptied
    if moved_spect_paths:
        unique_parents = set(
            [moved_spect.parent for moved_spect in moved_spect_paths]
        )
        for parent in unique_parents:
            if len(list(parent.iterdir())) < 1:
                shutil.rmtree(parent)
